class Error(Exception):
  def __init__(self, message = None):
    super(Error, self).__init__()

    self.message = message or ''

  def __str__(self):
    return self.message

class InvalidResourceError(Error):
  pass

class AccessViolationError(Error):
  pass

class AssemblerError(Error):
  def __init__(self, location = None, message = None, line = None, info = None):
    self.location = location
    self.message  = message
    self.line     = line
    self.info     = info

    self.create_text()

    super(AssemblerError, self).__init__(message = self.text[0])

  def create_text(self):
    text = ['{coor}: {message}'.format(coor = str(self.location), message = self.message)]

    if self.line is not None:
      text.append(self.line)

    if self.location.column is not None:
      text.append(' ' * self.location.column + '^')

    self.text = text

  def log(self, logger):
    logger('')
    for line in self.text:
      logger(line)

class TooManyLabelsError(AssemblerError):
  def __init__(self, **kwargs):
    super(TooManyLabelsError, self).__init__(message = 'Too many consecutive labels', **kwargs)

class UnknownPatternError(AssemblerError):
  def __init__(self, **kwargs):
    super(UnknownPatternError, self).__init__(message = 'Unknown pattern: "{info}"'.format(**kwargs), **kwargs)

class IncompleteDirectiveError(AssemblerError):
  def __init__(self, **kwargs):
    super(IncompleteDirectiveError, self).__init__(message = 'Incomplete directive: {info}'.format(**kwargs), **kwargs)

class UnknownFileError(AssemblerError):
  def __init__(self, **kwargs):
    super(UnknownFileError, self).__init__(message = 'Unknown file: {info}'.format(**kwargs), **kwargs)

class DisassembleMismatchError(AssemblerError):
  def __init__(self, **kwargs):
    super(DisassembleMismatchError, self).__init__(message = 'Disassembled instruction does not match input: {info}'.format(**kwargs), **kwargs)

class UnalignedJumpTargetError(AssemblerError):
  def __init__(self, **kwargs):
    super(UnalignedJumpTargetError, self).__init__(message = 'Jump destination address is not 4-byte aligned: {info}'.format(**kwargs), **kwargs)

class EncodingLargeValueError(AssemblerError):
  def __init__(self, **kwargs):
    super(EncodingLargeValueError, self).__init__(message = 'Value cannot fit into field: {info}'.format(**kwargs), **kwargs)

class IncompatibleLinkerFlagsError(Error):
  pass

class UnknownSymbolError(Error):
  pass



class IMachineWorker(object):
  """
  Base class for objects that provide pluggable service to others.
  """

  def boot(self, *args):
    """
    Prepare for providing the service. After this call, it may be requested
    by others.
    """

    pass

  def run(self):
    """
    Called by reactor's loop when this object is enqueued as a reactor task.
    """

    pass

  def suspend(self):
    """
    Suspend service. Object should somehow conserve its internal state, its
    service will not be used until the next call of ``wake_up`` method.
    """

    pass

  def wake_up(self):
    """
    Wake up service. In this method, object should restore its internal state,
    and after this call its service can be requested by others again.
    """

    pass

  def die(self, exc):
    """
    Exceptional state requires immediate termination of service. Probably no
    object will ever have need to call others' ``die`` method, it's intended
    for internal use only.
    """

    pass

  def halt(self):
    """
    Terminate service. It will never be requested again, object can destroy
    its internal state, and free allocated resources.
    """

    pass


class IReactorTask(object):
  """
  Base class for all reactor tasks.
  """

  def run(self):
    """
    This method is called by reactor to perform task's actions.
    """

    pass

class ISnapshotable(object):
  """
  Base class for objects that can be saved into a snapshot.
  """

  def save_state(self, parent):
    """
    Create state of the object, and attach it to a parent snapshot node.

    :param ducky.interfaces.ISnapshotable parent: parent snapshot node.
    """

    pass

  def load_state(self, state):
    """
    Restore state of the object.

    :param ducky.snapshot.SnapshotNode state: snapshot node containing saved state.
    """

    pass


class IVirtualInterrupt(object):
  def __init__(self, machine):
    super(IVirtualInterrupt, self).__init__()

    self.machine = machine

  def run(self, core):
    pass



"""
Streams represent basic IO objects, used by devices for reading or writing
(streams) of data.

``Stream`` object encapsulates an actual IO object - ``file``-like stream,
raw file descriptor, or even something completely different. ``Stream`` classes
then provide basic IO methods for moving data to and from stream, shielding
user from implementation details, like Python2/Python3 differencies.
"""

import abc
import errno
import fcntl
import io
import os
import termios
import tty

from six import PY2, integer_types, string_types, add_metaclass

from .errors import InvalidResourceError
from .util import isfile

def fd_blocking(fd, block = None):
  """
  Query or set blocking mode of file descriptor.

  :type int fd: file descriptor to manipulate.
  :type bool block: if set, method will set blocking mode of file descriptor
    accordingly: ``True`` means blocking, ``False`` non-blocking mode. If not
    set, the current setting will be returned. ``None`` by default.
  :rtype: bool
  :returns: if ``block`` is ``None``, current setting of blocking mode is
    returned - ``True`` for blocking, ``False`` for non-blocking. Othwerwise,
    function returns nothing.
  """

  flags = fcntl.fcntl(fd, fcntl.F_GETFL)

  if block is None:
    return (flags & os.O_NONBLOCK) == 0

  if block is True:
    flags &= ~os.O_NONBLOCK

  else:
    flags |= os.O_NONBLOCK

  fcntl.fcntl(fd, fcntl.F_SETFL, flags)

@add_metaclass(abc.ABCMeta)
class Stream(object):
  """
  Abstract base class of all streams.

  :param machine: parent :py:class`ducky.machine.Machine` object.
  :param desc: description of stream. This is a short, string representation
    of the stream.
  :param stream: ``file``-like stream that provides IO method (``read()`` or
    ``write``). If it is set, it is preferred IO object.
  :param int fd: raw file descriptor. ``stream`` takes precedence, otherwise this
    file descriptor is used.
  :param bool close: if ``True``, and if ``stream`` has a ``close()`` method, stream
    will provide ``close()`` method that will close the underlaying ``file``-like
    object. ``True`` by default.
  :param bool allow_close: if not ``True``, stream's ``close()`` method will *not*
    close underlying IO resource. ``True`` by default.
  """

  def __init__(self, machine, desc, stream = None, fd = None, close = True, allow_close = True):
    if stream is None and fd is None:
      raise InvalidResourceError('Stream "%s" must have stream object or raw file descriptor.' % desc)

    self.logger = machine.LOGGER
    self.DEBUG = machine.LOGGER.debug

    self.desc = desc
    self.stream = stream
    self.fd = fd

    self._raw_read = self._raw_read_stream if stream is not None else self._raw_read_fd
    self._raw_write = self._raw_write_stream if stream is not None else self._raw_write_fd

    self.allow_close = allow_close

    self._close = None
    if close is True and stream is not None and hasattr(stream, 'close'):
      self._close = stream.close

  def __repr__(self):
    return '<%s %s>' % (self.__class__.__name__, self.desc)

  def has_fd(self):
    """
    Check if stream has raw file descriptor. File descriptors can be checked
    for IO availability by reactor's polling task.

    :rtype: bool
    :returns: ``True`` when stream has file descriptor.
    """

    return self.fd is not None

  def has_poll_support(self):
    """
    Streams that can polled for data should return ``True``.

    :rtype: bool
    """

    # For most common case, if the stream has file descriptor set, it can be polled.
    return self.has_fd()

  def register_with_reactor(self, reactor, **kwargs):
    """
    Called by owner to register the stream with reactor's polling service.

    See :py:meth:`ducky.reactor.Reactor.add_fd` for keyword arguments.

    :param ducky.reactor.Reactor reactor: reactor instance to register with.
    """

    reactor.add_fd(self.fd, **kwargs)

  def unregister_with_reactor(self, reactor):
    """
    Called by owner to unregister the stream with reactor's polling service,
    e.g. when stream is about to be closed.

    :param ducky.reactor.Reactor reactor: reactor instance to unregister from.
    """

    reactor.remove_fd(self.fd)

  def _raw_read_stream(self, size = None):
    self.DEBUG('%s._raw_read_stream: size=%s', self.__class__.__name__, size)

    size = size or 0
    return self.stream.read(size)

  def _raw_read_fd(self, size = None):
    self.DEBUG('%s._raw_read_fd: size=%s', self.__class__.__name__, size)

    size = size or io.DEFAULT_BUFFER_SIZE
    return os.read(self.fd, size)

  def _raw_write_stream(self, data):
    self.DEBUG('%s._raw_write_stream: data="%s", len=%s, type=%s', self.__class__.__name__, data, len(data), type(data))

    remaining_chars = len(data)

    while remaining_chars > 0:
      try:
        self.stream.write(data)
        return

      except io.BlockingIOError as e:
        remaining_chars -= e.characters_written
        continue

      except EnvironmentError as e:
        # Resource temporarily unavailable
        if e.errno == 11:
          continue

        raise e

  def _raw_write_fd(self, data):
    self.DEBUG('%s._raw_write_fd: data="%s", len=%s, type=%s', self.__class__.__name__, data, len(data), type(data))

    os.write(self.fd, data)

  @abc.abstractmethod
  def read(self, size = None):
    """
    Read data from stream.

    :param int size: if set, read at maximum ``size`` bytes.
    :rtype: ``bytearray`` (Python2), ``bytes`` (Python3)
    :returns: read data, of maximum lenght of ``size``, ``None`` when there are
      no available data, or empty string in case of EOF.
    """

    raise NotImplementedError('%s does not implement read method' % self.__class__.__name__)

  @abc.abstractmethod
  def write(self, buff):
    """
    Write data to stream.

    :param bytearray buff: data to write. ``bytearray`` (Python2), ``bytes`` (Python3)
    """

    raise NotImplementedError('%s does not implement write method' % self.__class__.__name__)

  def close(self):
    """
    This method will close the stream. If ``allow_close`` flag is not set
    to ``True``, nothing will happen. If the stream wasn't created with ``close``
    set to ``True``, nothing will happen. If the wrapped IO resource does not
    have ``close()`` method, nothing will happen.
    """

    self.DEBUG('%s.close: allow_close=%s, _close=%s', self.__class__.__name__, self.allow_close, self._close)

    if not self.allow_close:
      self.DEBUG('%s.close: not allowed', self.__class__.__name__)
      return

    if self._close is not None:
      self._close()

class InputStream(Stream):
  if PY2:
    def read(self, size = None):
      self.DEBUG('%s.read: size=%s', self.__class__.__name__, size)

      buff = self._raw_read(size = size)
      if not buff:
        return bytearray([])

      return bytearray([ord(c) for c in buff])

  else:
    def read(self, size = None):
      self.DEBUG('%s.read: size=%s', self.__class__.__name__, size)

      buff = self._raw_read(size = size)

      return buff

  def write(self, b):
    raise NotImplementedError('%s does not implement write method' % self.__class__.__name__)

  @staticmethod
  def create(machine, desc):
    machine.LOGGER.debug('InputStream.create: desc=%s', desc)

    if isfile(desc):
      return FileInputStream(machine, desc)

    if hasattr(desc, 'read'):
      return MethodInputStream(machine, desc)

    if hasattr(desc, 'fileno'):
      return FDInputStream(machine, desc.fileno())

    if isinstance(desc, integer_types):
      return FDInputStream(machine, desc)

    if desc == '<stdin>':
      return StdinStream(machine)

    if isinstance(desc, string_types):
      return FileInputStream(machine, open(desc, 'rb'))

    raise InvalidResourceError('Unknown stream description: desc=%s' % desc)

class FileInputStream(InputStream):
  def __init__(self, machine, f, **kwargs):
    super(FileInputStream, self).__init__(machine, '<file %s>' % f.name, stream = f, fd = f.fileno())

class MethodInputStream(InputStream):
  def __init__(self, machine, desc, **kwargs):
    super(MethodInputStream, self).__init__(machine, repr(desc), stream = desc)

class FDInputStream(InputStream):
  def __init__(self, machine, fd, **kwargs):
    super(FDInputStream, self).__init__(machine, '<fd %s>' % fd, fd = fd)

class StdinStream(InputStream):
  def __init__(self, machine, **kwargs):
    DEBUG = machine.LOGGER.debug

    self.old_termios = None

    stream = machine.stdin.buffer if hasattr(machine.stdin, 'buffer') else machine.stdin
    fd = machine.stdin.fileno() if hasattr(machine.stdin, 'fileno') else None

    if fd is not None:
      DEBUG('%s.__init__: re-pack <stdin> fd as a new stream to avoid colisions', self.__class__.__name__)

      stream = os.fdopen(fd, 'rb', 0)
      fd = stream.fileno()

      DEBUG('%s.__init__: set <stdin> fd to non-blocking mode', self.__class__.__name__)

      fd_blocking(fd, block = False)

      try:
        self.old_termios = termios.tcgetattr(fd)
        tty.setcbreak(fd)

      except termios.error as e:
        if e.args[0] != errno.ENOTTY:
          raise e

      DEBUG('%s.__init__: stream=%r, fd=%r', self.__class__.__name__, stream, fd)

    super(StdinStream, self).__init__(machine, '<stdin>', stream = stream, fd = fd, close = False, **kwargs)

  def close(self):
    self.DEBUG('%s.close', self.__class__.__name__)

    if self.old_termios is not None:
      termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_termios)

  def get_selectee(self):
    return self.stream

class OutputStream(Stream):
  def read(self, size = None):
    raise NotImplementedError('%s does not implement read method' % self.__class__.__name__)

  if PY2:
    def write(self, buff):
      self.DEBUG('%s.write: buff=%s', self.__class__.__name__, buff)

      self._raw_write(''.join([chr(b) for b in buff]))

  else:
    def write(self, buff):
      self.DEBUG('%s.write: buff=%s', self.__class__.__name__, buff)

      self._raw_write(bytes(buff))

  def flush(self):
    if self.stream is not None and hasattr(self.stream, 'flush'):
      self.stream.flush()

  @staticmethod
  def create(machine, desc):
    machine.LOGGER.debug('OutputStream.create: desc=%s', desc)

    if isfile(desc):
      return FileOutputStream(machine, desc)

    if hasattr(desc, 'write'):
      return MethodOutputStream(machine, desc)

    if hasattr(desc, 'fileno'):
      return FDOutputStream(machine, desc.fileno())

    if isinstance(desc, integer_types):
      return FDOutputStream(machine, desc)

    if desc == '<stdout>':
      return StdoutStream(machine)

    if desc == '<stderr>':
      return StderrStream(machine)

    if isinstance(desc, string_types):
      return FileOutputStream(machine, open(desc, 'wb'))

    raise InvalidResourceError('Unknown stream description: desc=%s' % desc)

class FileOutputStream(OutputStream):
  def __init__(self, machine, f, **kwargs):
    super(FileOutputStream, self).__init__(machine, '<file %s>' % f.name, stream = f, fd = f.fileno())

class FDOutputStream(OutputStream):
  def __init__(self, machine, fd, **kwargs):
    super(FDOutputStream, self).__init__(machine, '<fd %s>' % fd, fd = fd)

class MethodOutputStream(OutputStream):
  def __init__(self, machine, desc, **kwargs):
    super(MethodOutputStream, self).__init__(machine, repr(desc), stream = desc)

class StdoutStream(OutputStream):
  def __init__(self, machine):
    stream = machine.stdout.buffer if hasattr(machine.stdout, 'buffer') else machine.stdout
    fd = machine.stdout.fileno() if hasattr(machine.stdout, 'fileno') else None

    super(StdoutStream, self).__init__(machine, '<stdout>', stream = stream, fd = fd, close = False)

class StderrStream(OutputStream):
  def __init__(self, machine):
    stream = machine.stderr.buffer if hasattr(machine.stderr, 'buffer') else machine.stderr
    fd = machine.stderr.fileno() if hasattr(machine.stderr, 'fileno') else None

    super(StderrStream, self).__init__(machine, '<stderr>', stream = stream, fd = fd, close = False)



from . import ASTVisitor

class ASTVisualiseVisitor(ASTVisitor):
  def __init__(self, *args, **kwargs):
    super(ASTVisualiseVisitor, self).__init__(*args, **kwargs)

    self.diag = None
    self.node_indices = []

  def get_index(self, node):
    try:
      return self.node_indices.index(node)

    except ValueError:
      self.node_indices.append(node)
      return self.node_indices.index(node)

  def generic_visit(self, node, shape = 'box'):
    self.diag.append('%i [numbered = %i, label = "%s", shape = %s];' % (self.get_index(node), self.get_index(node), node.__class__.__name__, shape))

    for c_name, c in node.children():
      self.visit(c)
      self.diag.append('%i -> %i [label = "%s"];' % (self.get_index(node), self.get_index(c), c_name))

  def visit_BinaryOp(self, node, shape = 'box'):
    self.diag.append('%i [numbered = %i, label = "BinaryOp\n%s", shape = %s];' % (self.get_index(node), self.get_index(node), node.op, shape))

    self.visit(node.left)
    self.diag.append('%i -> %i;' % (self.get_index(node), self.get_index(node.left)))

    self.visit(node.right)
    self.diag.append('%i -> %i;' % (self.get_index(node), self.get_index(node.right)))

  def visit_Constant(self, node, shape = 'box'):
    self.diag.append('%i [numbered = %i, label = "Constant (%s)", shape = %s];' % (self.get_index(node), self.get_index(node), node.type, shape))

  def visit_ID(self, node, shape = 'box'):
    self.diag.append('%i [numbered = %i, label = "ID\n%s", shape = %s];' % (self.get_index(node), self.get_index(node), node.name, shape))

  def visit_If(self, node):
    self.generic_visit(node, shape = 'diamond')

  def visit_UnaryOp(self, node, shape = 'box'):
    self.diag.append('%i [numbered = %i, label = "UnaryOp\n%s", shape = %s];' % (self.get_index(node), self.get_index(node), node.op, shape))

    self.visit(node.expr)
    self.diag.append('%i -> %i;' % (self.get_index(node), self.get_index(node.expr)))

  def visit_While(self, node):
    self.diag.append('%i [numbered = %i, label = "%s"];' % (self.get_index(node), self.get_index(node), node.__class__.__name__))

    self.visit(node.cond, shape = 'diamond')
    self.diag.append('%i -> %i [label = "cond"];' % (self.get_index(node), self.get_index(node.cond)))

    self.visit(node.stmt)
    self.diag.append('%i -> %i [label = "body"];' % (self.get_index(node), self.get_index(node.stmt)))

  def visit_FileAST(self, node):
    self.DEBUG(self.log_prefix + 'AST visualise pass')
    self.DOWN()

    self.diag = [
      'blockdiag {',
      'orientation = portrait;',
      'node_width = 256;',
      'node_height = 60;',
    ]

    self.generic_visit(node)

    self.diag.append('}')

    with open('ast.diag', 'w') as f:
      f.write('\n'.join(self.diag))

    self.UP()

from . import AST_PASSES
AST_PASSES['ast-visualise'] = ASTVisualiseVisitor



"""
Virtual machine debugging tools - break points, watch points, etc.

Create "point" that's triggered when a condition is satisfied (e.g.
processor executes instruction on specified address, memory at
specified address was modified, etc. Then, create "action" (e.g.
suspend core), and bind both pieces together - when point gets
triggered, execute list of actions.
"""

import enum
import logging

from six import itervalues

from .interfaces import IVirtualInterrupt
from .devices import IRQList, VIRTUAL_INTERRUPTS
from .util import str2int, UINT8_FMT, UINT16_FMT, UINT32_FMT
from .errors import InvalidResourceError

class Action(object):
  """
  Base class of all debugging actions.

  :param logging.Logger logger: logger instance used for logging.
  """

  def __init__(self, logger):
    self.logger = logger

  def act(self, core, point):
    """
    This method is called when "action" is executed. Implement it in child
    classes to give child actions a functionality.

    :param ducky.cpu.CPUCore core: CPU core where point was triggered.
    :param ducky.debugging.Point point: point that was triggered.
    """

    raise NotImplementedError()

  def __repr__(self):
    raise NotImplementedError()

class Point(object):
  """
  Base class of all debugging points.

  :param ducky.debugging.DebuggingSet debugging_set: debugging set this point belongs to.
  :param bool active: if not ``True``, point is not active and will not trigger.
  :param int countdown: if greater than zero, point has to trigger ``countdown`` times
    before its actions are executed for the first time.
  """

  def __init__(self, debugging_set, active = True, countdown = 0):
    super(Point, self).__init__()

    self.active = active
    self.countdown = countdown
    self.debugging_set = debugging_set

    self.actions = []

  def is_triggered(self, core, *args, **kwargs):
    """
    Test point's condition.

    :param ducky.cpu.CPUCore core: core requesting the test.
    :rtype: bool
    :returns: ``True`` if condition is satisfied.
    """

    raise NotImplementedError()

  def __repr__(self):
    raise NotImplementedError()

class SuspendCoreAction(Action):
  """
  If executed, this action will suspend the CPU core that triggered its parent
  point.
  """

  def act(self, core, point):
    core.suspend()

  def __repr__(self):
    return '<SuspendCoreAction>'

  @staticmethod
  def create_from_config(debugging_set, config, section):
    return SuspendCoreAction(debugging_set.core.LOGGER)

class LogValueAction(Action):
  """
  This is the base class for actions that log a numerical values.

  :param logging.Logger logger: logger instance used for logging.
  :param int size: size of logged number, in bytes.
  """

  def __init__(self, logger, size):
    super(LogValueAction, self).__init__(logger)

    self.size = size

    if size == 4:
      self.formatter = UINT32_FMT

    elif size == 2:
      self.formatter = UINT16_FMT

    else:
      self.formatter = UINT8_FMT

  def get_values(self, core, point):
    """
    Prepare dictionary with values for message that will be shown to the user.

    :param ducky.cpu.CPUCore core: core point was triggered on.
    :param ducky.debugging.Point point: triggered point.
    :rtype: dict
    :returns: dictionary that will be passed to message ``format()`` method.
    """

    raise NotImplementedError()

  def get_message(self, core, point):
    """
    Return message that, formatted with output of ``get_values()``, will be
    shown to user.

    :param ducky.cpu.CPUCore core: core point was triggered on.
    :param ducky.debugging.Point point: triggered point.
    :rtype: string
    :returns: information message.
    """

    raise NotImplementedError()

  def act(self, core, point):
    data = self.get_values(core, point)
    data.update({
      'watchpoint': repr(point),
      'ip':         UINT32_FMT(core.registers.ip.value),
      'value':      self.formatter(data['value'])
    })

    self.logger.info(self.get_message(core, point).format(**data))

class LogMemoryContentAction(LogValueAction):
  """
  When triggered, logs content of a specified location in memory.

  :param logging.Logger logger: logger instance used for logging.
  :param u32_t address: memory location.
  :param int size: size of logged number, in bytes.
  """

  def __init__(self, logger, address, size):
    super(LogMemoryContentAction, self).__init__(logger, size)

    self.address = address

    if self.size == 4:
      self.reader = 'read_u32'

    elif self.size == 2:
      self.reader = 'read_u16'

    else:
      self.reader = 'read_u8'

  def __repr__(self):
    return '<LogMemoryContentAction: address=%s, size=%s>' % (UINT32_FMT(self.address), self.size)

  def get_values(self, core, point):
    reader = getattr(core.mmu.memory, self.reader)
    return {
      'address': UINT32_FMT(self.address),
      'value':   reader(self.address),
    }

  def get_message(self, core, point):
    return 'memory: IP={ip}, {address}={value}'

  @staticmethod
  def create_from_config(debugging_set, config, section):
    _get, _getbool, _getint = config.create_getters(section)

    return LogMemoryContentAction(debugging_set.core.LOGGER, _getint('address'), _getint('size', 4))

class LogRegisterContentAction(LogValueAction):
  """
  When triggered, logs content of a specified register.

  :param logging.Logger logger: logger instance used for logging.
  :param list registers: list of register names.
  """

  def __init__(self, logger, registers):
    super(LogRegisterContentAction, self).__init__(logger, 4)

    self.registers = [r.strip() for r in registers.split(',')]
    self.formatter = UINT32_FMT

  def __repr__(self):
    return '<LogRegisterContentAction: registers=%s>' % ','.join(self.registers)

  def get_values(self, core, point):
    values = {'value': 0}
    for r in self.registers:
      values[r] = r
      values[r + '_value'] = UINT32_FMT(getattr(core.registers, r).value)

    return values

  def get_message(self, core, point):
    return 'register: IP={ip}, %s' % ', '.join(['{%s}={%s_value}' % (r, r) for r in self.registers])

  @staticmethod
  def create_from_config(debugging_set, config, section):
    _get, _getbool, _getint = config.create_getters(section)

    return LogRegisterContentAction(debugging_set.core.LOGGER, _get('registers'))

class BreakPoint(Point):
  def __init__(self, debugging_set, ip, *args, **kwargs):
    super(BreakPoint, self).__init__(debugging_set, *args, **kwargs)

    self.ip = ip

  def is_triggered(self, core):
    core.DEBUG('core IP=%s, self IP=%s', core.IP().value, self.ip)

    return core.IP().value == self.ip

  def __repr__(self):
    return '<BreakPoint: IP=%s>' % UINT32_FMT(self.ip)

  @staticmethod
  def create_from_config(debugging_set, config, section):
    _get, _getbool, _getint = config.create_getters(section)

    return BreakPoint(debugging_set, _getint('address'), active = _getbool('active', True), countdown = _getint('countdown', 0))

class MemoryWatchPoint(Point):
  def __init__(self, debugging_set, address, read, *args, **kwargs):
    super(MemoryWatchPoint, self).__init__(debugging_set, *args, **kwargs)

    self.address = address
    self.read = read

  def is_triggered(self, core, address = None, read = None):
    core.DEBUG('%s.is_triggered: address=%s, read=%s, self.address=%s, self.read=%s', self.__class__.__name__, UINT32_FMT(address), read, UINT32_FMT(self.address), self.read)

    if self.read is None:
      return address == self.address

    if self.read != read:
      return False

    return address == self.address

  def __repr__(self):
    return '<MemoryWatchPoint: address=%s>' % UINT32_FMT(self.address)

  @staticmethod
  def create_from_config(debugging_set, config, section):
    _get, _getbool, _getint = config.create_getters(section)

    return MemoryWatchPoint(debugging_set, _getint('address'), _getbool('read', None), active = _getbool('active', True), countdown = _getint('countdown', 0))

class DebuggingSet(object):
  def __init__(self, core):
    super(DebuggingSet, self).__init__()

    self.core = core
    self.points = []

    self.triggered_points = []

    C = core.cpu.machine.console

    console_commands = [
      ('bp-list', cmd_bp_list),
      ('bp-break', cmd_bp_add_breakpoint),
      ('bp-mwatch', cmd_bp_add_memory_watchpoint),
      ('bp-active', cmd_bp_active)
    ]

    for name, handler in console_commands:
      if C.is_registered_command(name):
        continue

      C.register_command(name, handler)

    for chain in ('step', 'memory'):
      setattr(self, 'triggered_%s' % chain, [])

      for stage in ('pre', 'post'):
        setattr(self, 'chain_%s_%s' % (stage, chain), [])

  def add_point(self, p, chain):
    self.core.DEBUG('adding point %s to chain %s', p, chain)

    getattr(self, 'chain_' + chain.replace('-', '_')).append(p)

  def remove_point(self, p, chain):
    self.core.DEBUG('removing point %s from chain %s', p, chain)

    getattr(self, 'chain_' + chain.replace('-', '_')).remove(p)

  def __check_chain(self, stage, chain, clean_triggered = False, *args, **kwargs):
    D = self.core.DEBUG

    D('__check_chain: stage=%s, chain=%s, clean_triggered=%s', stage, chain, clean_triggered)

    triggered = getattr(self, 'triggered_' + chain)
    chain = getattr(self, 'chain_%s_%s' % (stage, chain))

    D('__check_chain: before check: chain=%s, triggered=%s', str(chain), str(triggered))

    triggered_in_loop = 0

    for p in chain:
      D(repr(p))

      if not p.active:
        D('inactive, not evaluating')
        continue

      if not p.is_triggered(self.core, *args, **kwargs):
        D('not triggered, skipping')
        continue

      if p in triggered:
        D('already triggered by this step, ignore')
        continue

      if p.countdown > 0:
        p.countdown -= 1

      if p.countdown != 0:
        D('countdown %i, skip for now', p.countdown)
        continue

      self.core.INFO('Breakpoint triggered: %s', p)
      triggered.append(p)
      triggered_in_loop += 1

      for action in p.actions:
        action.act(self.core, p)

    D('__check_chain: after check: chain=%s, triggered=%s', str(chain), str(triggered))

    if clean_triggered is True:
      triggered[:] = []

    D('__check_chain: after cleanup: chain=%s, triggered=%s', str(chain), str(triggered))

    return triggered_in_loop > 0

  def pre_step(self):
    return self.__check_chain('pre', 'step')

  def post_step(self):
    return self.__check_chain('post', 'step', clean_triggered = True)

  def pre_memory(self, address = None, read = None):
    return self.__check_chain('pre', 'memory', address = address, read = read)

  def post_memory(self, address = None, read = None):
    return self.__check_chain('post', 'memory', clean_triggered = True, address = address, read = read)

def cmd_bp_list(console, cmd):
  """
  List existing breakpoints
  """

  points = [
    ['Point', 'Active', 'Countdown', 'Core']
  ]

  for point in itervalues(Point.points):
    points.append([
      repr(point),
      '*' if point.active else '',
      point.countdown,
      point.debugging_set.core.cpuid_prefix,
    ])

  console.table(points)

def cmd_bp_add_breakpoint(console, cmd):
  """
  Create new breakpoint: bp-break <#cpuid:#coreid> <address> [active] [countdown]
  """

  try:
    core = console.master.machine.core(cmd[1])

  except InvalidResourceError:
    console.write('go away')
    return

  ip = str2int(cmd[2])
  active = True if len(cmd) >= 3 and cmd[2] == 'yes' else False
  countdown = str2int(cmd[3]) if len(cmd) >= 4 else 0

  core.init_debug_set()
  point = core.debug.create_point(BreakPoint, ip, active = active, countdown = countdown)

  console.writeln('# OK: %s', point)

def cmd_bp_add_memory_watchpoint(console, cmd):
  """
  Create new memory watchpoint: bp-mwatch <#cpuid:#coreid> <address> [rw] [active] [countdown]'
  """

  try:
    core = console.master.machine.core(cmd[1])

  except InvalidResourceError:
    console.write('go away')
    return

  address = str2int(cmd[2])
  access = cmd[3] if len(cmd) >= 4 else 'r'
  active = True if len(cmd) >= 5 and cmd[4] == 'yes' else False
  countdown = str2int(cmd[5]) if len(cmd) >= 6 else 0

  core.init_debug_set()
  point = core.debug.create_point(MemoryWatchPoint, address, access, active = active, countdown = countdown)

  console.writeln('# OK: %s', point)

def cmd_bp_remove(console, cmd):
  """
  Remove breakpoint: bp-remove <id>
  """

  point = Point.points.get(int(cmd[1]))
  if point is None:
    console.writeln('go away')
    return

  point.debugging_set.remove_point(point)

  console.writeln('# OK')

def cmd_bp_active(console, cmd):
  """
  Toggle "active" flag for a breakpoint: bp-active <id>
  """

  point = Point.points.get(int(cmd[1]))
  if point is None:
    console.writeln('go away')
    return

  point.active = not point.active

  console.writeln('# OK: %s', point)

class VMDebugOperationList(enum.Enum):
  LOGGER_VERBOSITY = 0

class VMVerbosityLevels(enum.Enum):
  DEBUG = 0
  INFO = 1
  WARNING = 2
  ERROR = 3

VERBOSITY_LEVEL_MAP = {
  VMVerbosityLevels.DEBUG.value:   logging.DEBUG,
  VMVerbosityLevels.INFO.value:    logging.INFO,
  VMVerbosityLevels.WARNING.value: logging.WARNING,
  VMVerbosityLevels.ERROR.value:   logging.ERROR
}

class VMDebugInterrupt(IVirtualInterrupt):
  def run(self, core):
    from .cpu.registers import Registers

    core.DEBUG('VMDebugInterrupt: triggered')

    op = core.REG(Registers.R00).value
    core.REG(Registers.R00).value = 0xFFFF

    if op == VMDebugOperationList.LOGGER_VERBOSITY.value:
      verbosity = core.REG(Registers.R01).value

      if verbosity not in VERBOSITY_LEVEL_MAP:
        core.WARN('VMDebugInterrupt: unknown verbosity level: %s', verbosity)
        return

      core.LOGGER.setLevel(VERBOSITY_LEVEL_MAP[verbosity])
      core.DEBUG('VMDebugInterrupt: setting verbosity to %s', verbosity)

    else:
      core.WARN('VMDebugInterrupt: unknown operation requested: %s', op)

VIRTUAL_INTERRUPTS[IRQList.VMDEBUG.value] = VMDebugInterrupt



"""
This module provides support for profiling the virtual machine (`machine`
profilers) and running programs (`code` profilers). Wrappers for python's
deterministic profilers, and simple statistical profiler of running code
are available for usage thoughout the ducky sources.

``cProfile`` is the prefered choice of machine profiling backend, with
``profile`` is used as a backup option.

Beware, each thread needs its own profiler instance. These instances can be
acquired from ProfilerStore class which handles saving their data when virtual
machine exits.

To simplify the usage of profilers in ducky sources in case when user does not
need profiling, I chose to provide classes with the same API but these classes
does not capture any data. These are called `dummy` profilers, as opposed to
the `real` ones. Both kinds mimic API of :py:class:`profile.Profile` - the
`real` machine profiler **is** :py:class:`profile.Profile` object.
"""

import collections
import os
import os.path

from six.moves import cPickle as pickle

try:
  from cProfile import Profile as RealMachineProfiler

except ImportError:
  from profile import Profile as RealMachineProfiler

class DummyCPUCoreProfiler(object):
  """
  Dummy code profiler class. Base class for all code profilers.

  :param ducky.cpu.CPUCore core: core this profiler captures data from.
  :param int frequency: sampling frequency, given as an instruction count.
  """

  def __init__(self, core, frequency = 17):
    super(DummyCPUCoreProfiler, self).__init__()

    self.core = core
    self.frequency = frequency

    self.enabled = False

  def enable(self):
    """
    Enable collection of profiling data.
    """

    self.enabled = True

  def disable(self):
    """
    Disable collection of profiling data.
    """

    self.enabled = False

  def take_sample(self):
    """
    Take a sample of current state of CPU core, and store any necessary data.
    """

    pass

  def create_stats(self):
    """
    Preprocess collected data before they can be printed, searched or saved.
    """

    pass

  def dump_stats(self, filename):
    """
    Save collected data into file.

    :param string filename: path to file.
    """

    pass

class ProfileRecord(object):
  def __init__(self):
    self.ip = None
    self.instruction_set_id = None
    self.count = 0

  def merge(self, other):
    assert self.ip is None or self.ip == other.ip

    self.ip = other.ip
    self.instruction_set_id = other.instruction_set_id
    self.count += other.count

class RealCPUCoreProfiler(DummyCPUCoreProfiler):
  """
  Real code profiler. This class actually does collect data.
  """

  def __init__(self, core):
    super(RealCPUCoreProfiler, self).__init__(core)

    self.data = []

  def take_sample(self):
    if self.enabled is not True:
      return

    if self.core.registers.cnt.value % self.frequency != 0:
      return

    self.data.append((self.core.instruction_set.instruction_set_id, self.core.current_ip))

  def dump_stats(self, filename):
    processed = collections.defaultdict(ProfileRecord)

    for inst_set, ip in self.data:
      processed[ip].ip = ip
      processed[ip].instruction_set_id = inst_set
      processed[ip].count += 1

    with open(filename, 'wb') as f:
      pickle.dump(processed, f)

class DummyMachineProfiler(object):
  """
  Dummy machine profiler. Does absolutely nothing.
  """

  def __init__(self, *args, **kwargs):
    super(DummyMachineProfiler, self).__init__()

  def enable(self):
    """
    Start collecting profiling data.
    """

    pass

  def disable(self):
    """
    Stop collecting profiling data.
    """

    pass

  def create_stats(self):
    """
    Preprocess collected data before they can be printed, searched or saved.
    """

    pass

  def dump_stats(self, filename):
    """
    Save collected data into file.

    :param string filename: path to file.
    """

    pass

class ProfilerStore(object):
  """
  This class manages profiler instances. When in need of a profiler (e.g. in a
  new thread) get one by calling proper method of ProfilerStore object.
  """

  def __init__(self):
    super(ProfilerStore, self).__init__()

    self.machine_profiler_class = DummyMachineProfiler
    self.core_profiler_class    = DummyCPUCoreProfiler

    self.profilers = []

  def enable_machine(self):
    """
    Each newly created virtual machine profiler will be the real one.
    """

    self.machine_profiler_class = RealMachineProfiler

  def enable_cpu(self):
    """
    Each newly created code profiler will be the real one.
    """

    self.core_profiler_class = RealCPUCoreProfiler

  def is_machine_enabled(self):
    """
    Returns ``True`` when virtual machine profiling is enabled.

    :rtype: bool
    """

    return self.machine_profiler_class == RealMachineProfiler

  def is_cpu_enabled(self):
    """
    Returns ``True`` when code profiling is enabled.

    :rtype: bool
    """

    return self.core_profiler_class == RealCPUCoreProfiler

  def get_machine_profiler(self):
    """
    Create and return new machine profiler.

    :returns: new machine profiler.
    """

    p = self.machine_profiler_class(builtins = False)

    self.profilers.append(p)

    return p

  def get_core_profiler(self, core):
    """
    Create new code profiler.

    :rtype: DummyCPUCoreProfiler
    """

    p = self.core_profiler_class(core)

    self.profilers.append(p)

    return p

  def save(self, logger, directory):
    """
    Save all captured data to files. Each created profiler stores its data in
    separate file.

    :param string directory: directory where all files are stored.
    """

    filename_pattern = os.path.join(directory, 'profiler-%s-%s-%i.dat')

    for index, profiler in enumerate(self.profilers):
      profiler.disable()
      profiler.create_stats()

      filename = filename_pattern % (os.getpid(), profiler.__class__.__name__, index)
      profiler.dump_stats(filename)

      logger.info('Profiling data: %s', filename)

#: Main profiler store
STORE = ProfilerStore()



import argparse
import logging
import optparse
import sys

from ..log import create_logger, StreamHandler

def setup_logger(stream = None, debug = False, quiet = None, verbose = None, default_loglevel = logging.INFO):
  stream = stream or sys.stdout

  logger = create_logger(handler = StreamHandler(stream = stream))
  logger.setLevel(logging.INFO)

  if debug:
    logger.setLevel(logging.DEBUG)

  else:
    levels = [logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL]
    quiet = quiet or 0
    verbose = verbose or 0

    level = levels.index(default_loglevel) + quiet - verbose
    level = max(0, level)
    level = min(4, level)

    logger.setLevel(levels[level])

  return logger

def add_common_options(parser):
  group = optparse.OptionGroup(parser, 'Tool verbosity')
  parser.add_option_group(group)

  group.add_option('-d', '--debug', dest = 'debug', action = 'store_true', default = False, help = 'Debug mode')
  group.add_option('-q', '--quiet', dest = 'quiet', action = 'count', default = 0, help = 'Decrease verbosity. This option can be used multiple times')
  group.add_option('-v', '--verbose', dest = 'verbose', action = 'count', default = 0, help = 'Increase verbosity. This option can be used multiple times')

def parse_options(parser, default_loglevel = logging.INFO, stream = None):
  stream = stream or sys.stdout

  if isinstance(parser, argparse.ArgumentParser):
    options = parser.parse_args()
  else:
    options, args = parser.parse_args()

  logger = setup_logger(stream = sys.stdout, debug = options.debug, quiet = options.quiet, verbose = options.verbose, default_loglevel = default_loglevel)

  from signal import signal, SIGPIPE, SIG_DFL
  signal(SIGPIPE, SIG_DFL)

  return options, logger



from pycparser import c_ast

from . import ASTOptVisitor, dump_node

class DSEVisitor(ASTOptVisitor):
  priority = 500

  def visit_Compound(self, node):
    #
    # Any statement after "return" in a block is effectively dead
    #
    if any(isinstance(stmt, c_ast.Return) for stmt in node.block_items):
      self.DEBUG(self.log_prefix + 'DCE: Return present in compound block, check for dead statements')
      self.DOWN()

      for stmt in reversed(node.block_items[:]):
        if isinstance(stmt, c_ast.Return):
          break

        self.DEBUG(self.log_prefix + 'removing %s', dump_node(stmt))
        node.block_items.remove(stmt)
        self.tree_modified = True

      self.generic_visit(node)

      self.UP()

from . import AST_PASSES
AST_PASSES['ast-dce'] = DSEVisitor



import os
import sys

from six import iteritems, PY2

def translate_buffer(logger, buffer, file_in, options):
  from ..cpu.assemble import AssemblerError, translate_buffer

  try:
    return translate_buffer(logger, buffer, mmapable_sections = options.mmapable_sections, writable_sections = options.writable_sections, filename = file_in, defines = options.defines, includes = options.includes, verify_disassemble = options.verify_disassemble)

  except AssemblerError as exc:
    exc.log(logger.error)
    sys.exit(1)


def encode_blob(logger, file_in, options):
  logger.debug('encode_blob: file_in=%s', file_in)

  from ..cpu.assemble import DataSection, SymbolsSection, IntSlot, Label, sizeof, BytesSlot, RelocSection, SourceLocation
  from ..mm.binary import SectionFlags

  section_name = '__' + os.path.split(file_in)[1].replace('.', '_').replace('-', '_')
  section = DataSection('.%s' % section_name, flags = SectionFlags.from_string(options.blob_flags.upper()))
  section.base = section.ptr = 0

  symtab = SymbolsSection('.symtab')
  symtab.base = symtab.ptr = 0

  reloc = RelocSection('.reloc')
  reloc.base = reloc.ptr = 0

  if options.mmapable_sections is True:
    section.flags.mmapable = True

  logger.debug('section: %s', section)

  with open(file_in, 'rb') as f_in:
    if PY2:
      data = bytearray([ord(c) for c in f_in.read()])

    else:
      data = f_in.read()

  var_size = IntSlot()
  var_size.name = Label('%s_size' % section_name, section, SourceLocation(filename = file_in, lineno = 0))
  var_size.value = len(data)
  var_size.section = section
  var_size.section_ptr = 0
  var_size.close()

  section.content += var_size.value
  symtab.content.append(var_size)

  var_content = BytesSlot()
  var_content.name = Label('%s_start' % section_name, section, SourceLocation(filename = file_in, lineno = 0))
  var_content.value = data
  var_content.section = section
  var_content.section_ptr = sizeof(var_size)
  var_content.close()

  section.content += var_content.value
  symtab.content.append(var_content)

  logger.debug('section: %s', section)

  return {section.name: section, '.symtab': symtab, '.reloc': reloc}

def save_object_file(logger, sections, file_out, options):
  if os.path.exists(file_out) and not options.force:
    logger.error('Output file %s already exists, use -f to force overwrite', file_out)
    sys.exit(1)

  from ..cpu.assemble import sizeof
  from ..mm.binary import File, SectionTypes, SymbolEntry, RelocEntry

  section_name_to_index = {}

  i = 0
  for s_name, section in list(iteritems(sections)):
    if section.type in (SectionTypes.SYMBOLS, SectionTypes.RELOC):
      continue

    if section.flags.bss and section.data_size > 0:
      continue

    if section.data_size > 0:
      continue

    del sections[s_name]

    section_name_to_index[s_name] = i
    i += 1

  with File.open(logger, file_out, 'w') as f_out:
    h_file = f_out.create_header()
    h_file.flags.mmapable = 1 if options.mmapable_sections else 0

    filenames = {}

    section_name_to_index = {}
    for i, s_name in enumerate(sections.keys()):
      section_name_to_index[s_name] = i

    for s_name, section in iteritems(sections):
      h_section = f_out.create_section()
      h_section.type = section.type
      h_section.items = section.items
      h_section.data_size = section.data_size
      h_section.file_size = section.file_size
      h_section.name = f_out.string_table.put_string(section.name)
      h_section.base = section.base

      if section.type == SectionTypes.SYMBOLS:
        symbol_entries = []

        for se in section.content:
          entry = SymbolEntry()
          symbol_entries.append(entry)

          entry.flags = se.flags.to_encoding()
          entry.name = f_out.string_table.put_string(se.name.name)
          entry.address = se.section_ptr
          entry.size = se.size
          entry.section = section_name_to_index[se.section.name]

          if se.location is not None:
            loc = se.location

            if loc.filename not in filenames:
              filenames[loc.filename] = f_out.string_table.put_string(loc.filename)
            entry.filename = filenames[loc.filename]

            if loc.lineno:
              entry.lineno = loc.lineno

          entry.type = se.symbol_type

        f_out.set_content(h_section, symbol_entries)
        h_section.data_size = h_section.file_size = sizeof(SymbolEntry()) * len(symbol_entries)

      elif section.type == SectionTypes.RELOC:
        reloc_entries = []

        for rs in section.content:
          entry = RelocEntry()
          reloc_entries.append(entry)

          entry.name = f_out.string_table.put_string(rs.name)
          entry.flags = rs.flags.to_encoding()
          entry.patch_section = section_name_to_index[rs.patch_section.name]
          entry.patch_address = rs.patch_address
          entry.patch_offset = rs.patch_offset or 0
          entry.patch_size = rs.patch_size
          entry.patch_add = rs.patch_add or 0

        f_out.set_content(h_section, reloc_entries)
        h_section.data_size = h_section.file_size = sizeof(RelocEntry()) * len(reloc_entries)

      else:
        h_section.flags = section.flags.to_encoding()
        f_out.set_content(h_section, section.content)

    h_section = f_out.create_section()
    h_section.type = SectionTypes.STRINGS
    h_section.name = f_out.string_table.put_string('.strings')

    f_out.save()

def main():
  import optparse
  from . import add_common_options, parse_options

  parser = optparse.OptionParser()
  add_common_options(parser)

  group = optparse.OptionGroup(parser, 'File options')
  parser.add_option_group(group)
  group.add_option('-i', dest = 'file_in', action = 'append', default = [], help = 'Input file')
  group.add_option('-o', dest = 'file_out', action = 'append', default = [], help = 'Output file')
  group.add_option('-f', dest = 'force', default = False, action = 'store_true', help = 'Force overwrite of the output file')

  group = optparse.OptionGroup(parser, 'Translation options')
  parser.add_option_group(group)
  group.add_option('-D', dest = 'defines', action = 'append', default = [], help = 'Define variable', metavar = 'VAR')
  group.add_option('-I', dest = 'includes', action = 'append', default = [], help = 'Add directory to list of include dirs', metavar = 'DIR')
  group.add_option('--verify-disassemble', dest = 'verify_disassemble', action = 'store_true', default = False, help = 'Verify that disassebler instructions match input text')

  group = optparse.OptionGroup(parser, 'Binary options')
  parser.add_option_group(group)
  group.add_option('-b', '--blob',              dest = 'blob',              action = 'store_true', default = False, help = 'Create object file wrapping a binary blob')
  group.add_option('-B', '--blob-flags',        dest = 'blob_flags',        action = 'store',      default = 'rl',  help = 'Flags of blob section')
  group.add_option('-m', '--mmapable-sections', dest = 'mmapable_sections', action = 'store_true', default = False, help = 'Create mmap\'able sections')
  group.add_option('-w', '--writable-sections', dest = 'writable_sections', action = 'store_true', default = False, help = '.text and other read-only sections will be marked as writable too')

  options, logger = parse_options(parser)

  if not options.file_in:
    parser.print_help()
    sys.exit(1)

  if len(options.file_out) and len(options.file_out) != len(options.file_in):
    logger.error('If specified, number of output files must be equal to number of input files')
    sys.exit(1)

  for file_in in options.file_in:
    with open(file_in, 'r') as f_in:
      buffer = f_in.read()

    if options.file_out:
      file_out = options.file_out.pop(0)

    else:
      file_out = os.path.splitext(file_in)[0] + '.o'

    if options.blob is True:
      sections = encode_blob(logger, file_in, options)

    else:
      sections = translate_buffer(logger, buffer, file_in, options)

    save_object_file(logger, sections, file_out, options)



import sys
import optparse
import string

from six import print_, iteritems
from functools import partial

from ..snapshot import CoreDumpFile
from ..mm import PAGE_SIZE, UINT32_FMT, PAGE_MASK, u32_t, u16_t, u8_t, UINT8_FMT, UINT16_FMT
from ..mm.binary import File, SectionTypes
from ..cpu import CoreFlags
from ..cpu.registers import Registers
from ..log import WHITE, GREEN
from ..util import str2int, align

def show_header(logger, state):
  state = state.get_child('machine')

  logger.info('=== Coredump header ===')
  logger.info('  # of CPUs:      %i', state.nr_cpus)
  logger.info('  # of CPU cores: %i', state.nr_cores)
  logger.info('')

def show_cores(logger, state):
  logger.info('=== Cores ===')

  def __reg(index):
    return UINT32_FMT(cs.registers[index])

  for cpu_state in state.get_child('machine').get_cpu_states():
    for cs in cpu_state.get_core_states():
      logger.info('Core #%i:#%i', cs.cpuid, cs.coreid)

      for i in range(0, Registers.REGISTER_SPECIAL, 4):
        regs = [(i + j) for j in range(0, 4) if (i + j) < Registers.REGISTER_SPECIAL]
        s = ['reg%02i=%s' % (reg, __reg(reg)) for reg in regs]
        logger.info('  %s', ' '.join(s))

      flags = CoreFlags.from_int(cs.flags)

      logger.info('  fp=%s    sp=%s    ip=%s', __reg(Registers.FP), __reg(Registers.SP), __reg(Registers.IP))
      logger.info('  flags=%s', flags.to_string())
      logger.info('  cnt=%i, alive=%s, running=%s, idle=%s, exit=%i', cs.registers[Registers.CNT], cs.alive, cs.running, cs.idle, cs.exit_code)
      logger.info('  ivt=%s', UINT32_FMT(cs.ivt_address))
      logger.info('  pt= %s, pt-enabled=%s', UINT32_FMT(cs.pt_address), cs.pt_enabled)

      logger.info('')

def show_memory(logger, state):
  logger.info('=== Memory ===')

  state = state.get_child('machine').get_child('memory')

  logger.info('  Size:             %s', state.size)
  logger.info('  # of dirty pages: %s', len(state.get_page_states()))
  logger.info('')

def show_pages(logger, state, empty_pages = False):
  logger.info('=== Memory pages ===')

  for pg in sorted(state.get_child('machine').get_child('memory').get_page_states(), key = lambda x: x.index):
    if not empty_pages and all(i == 0 for i in pg.content):
      continue

    CPR = 32
    pg_addr = pg.index * PAGE_SIZE

    logger.info('  Page #%8i   %s', pg.index, ' '.join(['%02X' % i for i in range(CPR)]))
    # logger.info('    Flags: %s', pg.flags.to_string())

    for i in range(0, 256 // CPR):
      s = []
      t = []

      for b in pg.content[CPR * i:CPR * (i + 1)]:
        c = '%02X' % b
        s.append(GREEN(c) if b == 0 else WHITE(c))

        c = chr(b)
        if c in string.printable[0:-5]:
          c = r'%' if c == '%' else (' ' + c)
          t.append(c)

        elif b == 10:
          t.append(r'\n')

        elif b == 13:
          t.append(r'\r')

        else:
          t.append('  ')

      logger.info('    ' + UINT32_FMT(pg_addr + i * CPR) + ':    ' + ' '.join(s))
      logger.info('                  ' + ' '.join(t))

    logger.info('')

def __load_forth_symbols(logger):
  symbols = {}

  with File.open(logger, 'forth/ducky-forth', 'r') as f:
    f.load()

    for i in range(0, f.get_header().sections):
      header, content = f.get_section(i)

      if header.type != SectionTypes.SYMBOLS:
        continue

      for index, entry in enumerate(content):
        symbols[entry.address] = f.string_table.get_string(entry.name)

  return symbols

def __read(state, cnt, address):
  pid = (address & PAGE_MASK) >> 8
  offset = address & 0xFF

  pg = [pg for pg in state.get_child('machine').get_child('memory').get_page_states() if pg.index == pid][0]

  if cnt == 1:
    return u8_t(pg.content[offset])

  if cnt == 2:
    return u16_t(pg.content[offset] | (pg.content[offset + 1] << 8))

  if cnt == 4:
    return u32_t(pg.content[offset] | (pg.content[offset + 1] << 8) | (pg.content[offset + 2] << 16) | (pg.content[offset + 3] << 24))

def __show_forth_word(logger, state, symbols, base_address, ending_addresses):
  __read_u8  = partial(__read, state, 1)
  __read_u16 = partial(__read, state, 2)
  __read_u32 = partial(__read, state, 4)

  namelen = __read_u8(base_address + 7).value

  code_address = align(4, base_address + 8 + namelen)

  logger.info('Base address: %s', UINT32_FMT(base_address))
  logger.info('Link:         %s', UINT32_FMT(__read_u32(base_address)))
  logger.info('CRC:          %s', UINT16_FMT(__read_u16(base_address + 4)))
  logger.info('flags:        %s', UINT8_FMT(__read_u8(base_address + 6)))
  logger.info('namelen:      %s', namelen)
  logger.info('name:         %s', ''.join([chr(__read_u8(base_address + 8 + i).value) for i in range(0, namelen)]))

  while True:
    code_token = __read_u32(code_address).value
    token_name = symbols.get(code_token, '<unknown>')

    logger.info('              %s - %s', UINT32_FMT(code_token), token_name)

    if code_token in ending_addresses:
      break

    if token_name.startswith('code_'):
      break

    code_address += 4

def show_forth_trace(logger, state):
  logger.info('=== FORTH call trace ===')

  bottom = (state.get_child('machine').get_cpu_states()[0].get_core_states()[0].registers[21] & 0xFFFFFF00) / PAGE_SIZE
  top    = 0xFFFFFA
  pages  = [pg for pg in state.get_child('machine').get_child('memory').get_page_states() if bottom <= pg.index < top]

  stack = sorted(pages, key = lambda x: x.index, reverse = True)

  symbols = __load_forth_symbols(logger)

  for pg in stack:
    for i in range(PAGE_SIZE, 0, -4):
      step = pg.content[i - 4] | (pg.content[i - 3] << 8) | (pg.content[i - 2] << 16) | (pg.content[i - 1] << 24)

      if step != 0 and step in symbols:
        logger.info('%s: %s', UINT32_FMT(step), symbols[step])

      else:
        logger.info('%s: %s', UINT32_FMT(step), UINT32_FMT(step))

def show_forth_word(logger, state, base_address):
  logger.info('=== FORTH word ===')

  symbols = __load_forth_symbols(logger)

  __show_forth_word(logger, state, symbols, base_address, [[k for k, v in iteritems(symbols) if v == 'EXIT'][0]])

def show_forth_dict(logger, state, last):
  logger.info('== FORTH dictionary ===')

  __read_u32 = partial(__read, state, 4)

  symbols = __load_forth_symbols(logger)

  base_address = __read_u32(last).value
  ending_addresses = [[k for k, v in iteritems(symbols) if v == 'EXIT'][0]]

  while base_address != 0x00000000:
    logger.info('')
    __show_forth_word(logger, state, symbols, base_address, ending_addresses)
    base_address = __read_u32(base_address).value

def main():
  from . import add_common_options, parse_options

  parser = optparse.OptionParser()
  add_common_options(parser)

  parser.add_option('-i', dest = 'file_in', default = None, help = 'Input file')

  parser.add_option('-H',         dest = 'header',   default = False, action = 'store_true', help = 'Show file header')
  parser.add_option('-C',         dest = 'cores',    default = False, action = 'store_true', help = 'Show cores')
  parser.add_option('-M',         dest = 'memory',   default = False, action = 'store_true', help = 'Show memory')
  parser.add_option('--pages',    dest = 'pages',    default = False, action = 'store_true', help = 'Show pages')
  parser.add_option('--empty-pages', dest = 'empty_pages', default = False, action = 'store_true', help = 'Show empty pages')
  parser.add_option('--forth-trace', dest = 'forth_trace', default = False, action = 'store_true', help = 'Show FORTH call trace')
  parser.add_option('--forth-word',  dest = 'forth_word',  default = None,  action = 'store',    help = 'Show FORTH word')
  parser.add_option('--forth-dict',  dest = 'forth_dict',  default = None,  action = 'store',      help = 'Show FORTH dictionary')
  parser.add_option('-a',         dest = 'all',      default = False, action = 'store_true', help = 'All of above')

  parser.add_option('-Q',         dest = 'queries',  default = [],    action = 'append',     help = 'Query snapshot')

  options, logger = parse_options(parser)

  if not options.file_in:
    parser.print_help()
    sys.exit(1)

  if options.all:
    options.header = options.cores = options.memory = options.pages = True

  logger.info('Input file: %s', options.file_in)

  with CoreDumpFile.open(logger, options.file_in, 'r') as f_in:
    state = f_in.load()

    if not options.queries:
      logger.info('')

      if options.header:
        show_header(logger, state)

      if options.cores:
        show_cores(logger, state)

      if options.memory:
        show_memory(logger, state)

      if options.pages:
        show_pages(logger, state, empty_pages = options.empty_pages)

      if options.forth_trace:
        show_forth_trace(logger, state)

      if options.forth_word:
        show_forth_word(logger, state, str2int(options.forth_word))

      if options.forth_dict:
        show_forth_dict(logger, state, str2int(options.forth_dict))

    else:
      for query in options.queries:
        print_(eval(query, {'STATE': state}), end = '')

if __name__ == '__main__':
  main()



import enum

from ..mm import u32_t, u64_t

class Registers(enum.IntEnum):
  # General purpose registers
  R00   =  0
  R01   =  1
  R02   =  2
  R03   =  3
  R04   =  4
  R05   =  5
  R06   =  6
  R07   =  7
  R08   =  8
  R09   =  9
  R10   = 10
  R11   = 11
  R12   = 12
  R13   = 13
  R14   = 14
  R15   = 15
  R16   = 16
  R17   = 17
  R18   = 18
  R19   = 19
  R20   = 20
  R21   = 21
  R22   = 22
  R23   = 23
  R24   = 24
  R25   = 25
  R26   = 26
  R27   = 27
  R28   = 28
  R29   = 29

  # Special registers
  FP    = 30  # Frame Pointer
  SP    = 31  # Stack Pointer

  # Inaccessible registers
  IP    = 32  # Instruction pointer
  CNT   = 33  # Instruction counter

  # First special register
  REGISTER_SPECIAL = 30

  # How many registers do we have? This many...
  REGISTER_COUNT = 34

PROTECTED_REGISTERS = [
  Registers.FP,
  Registers.IP,
  Registers.CNT
]

FLAGS = Registers.REGISTER_COUNT.value + 100

GENERAL_REGISTERS   = [r for r in Registers if r.value < Registers.REGISTER_SPECIAL.value]
RESETABLE_REGISTERS = [r for r in Registers if r.value < Registers.CNT]

REGISTER_NAMES = ['r{}'.format(r.value) for r in Registers if r.value < Registers.REGISTER_SPECIAL.value] + ['fp', 'sp', 'ip', 'cnt']

class RegisterSet(object):
  def __init__(self):
    super(RegisterSet, self).__init__()

    self.map = {}

    for register_name, register_id in zip(REGISTER_NAMES, Registers):
      register = u32_t(0) if register_name != 'cnt' else u64_t(0)

      setattr(self, register_name, register)
      self.map[register_name] = register
      self.map[register_id.value] = register



"""
Stack-based coprocessor, providing several arithmetic operations with "long"
numbers.

Coprocessor's instructions operates on a stack of (by default) 8 slots.
Operations to move values between math stack and registers/data stack are also
available.

In the following documentation several different data types are used:

 - ``int`` - standard `word`, 32-bit wide integer
 - ``long`` - long integer, 64-bit wide

Unless said otherwise, instruction takes its arguments from the stack, removing
the values in the process, and pushes the result - if any - back on the stack.
"""

import enum

from ...interfaces import ISnapshotable
from . import Coprocessor
from .. import CPUException
from ..instructions import InstructionSet, SIS, INSTRUCTION_SETS, Descriptor_R_R, ENCODE, REGISTER_NAMES
from ...mm import u32_t, i64_t, u64_t, UINT64_FMT
from ...snapshot import SnapshotNode

#: Number of available spots on the math stack.
STACK_DEPTH = 8

class MathCoprocessorState(SnapshotNode):
  """
  Snapshot node holding the state of math coprocessor.
  """

  def __init__(self):
    super(MathCoprocessorState, self).__init__('stack')

class EmptyMathStackError(CPUException):
  """
  Raised when operation expects at least one value on math stack but stack
  is empty.
  """

  def __init__(self, *args, **kwargs):
    super(EmptyMathStackError, self).__init__('Math stack is empty', *args, **kwargs)

class FullMathStackError(CPUException):
  """
  Raised when operation tries to put value on math stack but there is no empty
  spot available.
  """

  def __init__(self, *args, **kwargs):
    super(FullMathStackError, self).__init__('Math stack is full', *args, **kwargs)

class RegisterSet(ISnapshotable):
  """
  Math stack wrapping class. Provides basic push/pop access, and direct access
  to a top of the stack.

  :param ducky.cpu.CPUCore core: CPU core registers belong to.
  """

  def __init__(self, core):
    super(RegisterSet, self).__init__()

    self.core = core
    self.stack = []

    self.DEBUG = core.DEBUG

  def save_state(self, parent):
    self.DEBUG('RegisterSet.save_state')

    state = parent.add_child('math_coprocessor', MathCoprocessorState())

    state.stack = []
    for lr in self.stack:
      state.stack.append(int(lr.value))

  def load_state(self, state):
    self.DEBUG('RegisterSet.load_state')

    for lr in state.depth:
      self.stack.append(u64_t(lr))

  def push(self, v):
    """
    Push new value on top of the stack.

    :raises ducky.cpu.coprocessor.math_copro.FullMathStackError: if there is no space available on the stack.
    """

    self.DEBUG('%s.push: v=%s', self.__class__.__name__, UINT64_FMT(v))

    if len(self.stack) == STACK_DEPTH:
      raise FullMathStackError()

    self.stack.append(v)

  def pop(self):
    """
    Pop the top value from stack and return it.

    :raises ducky.cpu.coprocessor.math_copro.EmptyMathStackError: if there are no values on the stack.
    """

    self.DEBUG('%s.pop', self.__class__.__name__)

    if not self.stack:
      raise EmptyMathStackError()

    return self.stack.pop()

  def tos(self):
    """
    Return the top of the stack, without removing it from a stack.

    :raises ducky.cpu.coprocessor.math_copro.EmptyMathStackError: if there are no values on the stack.
    """

    self.DEBUG('%s.tos', self.__class__.__name__)

    if not self.stack:
      raise EmptyMathStackError()

    return self.stack[-1]

  def tos1(self):
    """
    Return the item below the top of the stack, without removing it from a stack.

    :raises ducky.cpu.coprocessor.math_copro.EmptyMathStackError: if there are no values on the stack.
    """

    self.DEBUG('%s.tos', self.__class__.__name__)

    if len(self.stack) < 2:
      raise EmptyMathStackError()

    return self.stack[-2]

class MathCoprocessor(ISnapshotable, Coprocessor):
  """
  Coprocessor itself, includes its register set ("math stack").

  :param ducky.cpu.CPUCore core: CPU core coprocessor belongs to.
  """

  def __init__(self, core, *args, **kwargs):
    super(MathCoprocessor, self).__init__(core, *args, **kwargs)

    self.registers = RegisterSet(core)

  def save_state(self, parent):
    self.registers.save_state(parent)

  def load_state(self, state):
    self.registers.load_state(state)

  def dump_stack(self):
    """
    Log content of the stack using parent's ``DEBUG`` method.
    """

    D = self.core.DEBUG

    D('Math stack:')

    for index, lr in enumerate(self.registers.stack):
      D('#%02i: %s', index, UINT64_FMT(lr.value))

    D('---')

  def sign_extend_with_push(self, i32):
    v = u64_t(i32)

    if i32 & 0x80000000:
      v.value |= 0xFFFFFFFF00000000

    self.registers.push(v)

  def extend_with_push(self, u32):
    self.registers.push(u64_t(u32))

#
# Instruction set
#
class MathCoprocessorOpcodes(enum.IntEnum):
  """
  Math coprocessor's instruction opcodes.
  """

  POPW    = 0
  POPUW   = 1
  PUSHW   = 2
  SAVEW   = 3
  LOADW   = 4
  LOADUW  = 5

  POP     = 6
  SAVE    = 7
  PUSH    = 8
  LOAD    = 9

  MULL    = 10
  DIVL    = 11
  MODL    = 12
  SYMDIVL = 13
  SYMMODL = 14
  UDIVL   = 15
  UMODL   = 16

  DUP     = 20
  DUP2    = 21
  SWP     = 22
  DROP    = 23

  INCL    = 30
  DECL    = 31
  ADDL    = 32

  SIS     = 63

class MathCoprocessorInstructionSet(InstructionSet):
  """
  Math coprocessor's instruction set.
  """

  instruction_set_id = 1

  opcodes = MathCoprocessorOpcodes

class Descriptor_MATH(Descriptor_R_R):
  operands = ''

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    pass

  @staticmethod
  def disassemble_operands(logger, inst):
    pass

class PUSHW(Descriptor_MATH):
  """
  Downsize TOS to ``int``, and push the result on the data stack.
  """

  mnemonic = 'pushw'
  opcode = MathCoprocessorOpcodes.PUSHW

  @staticmethod
  def execute(core, inst):
    core.raw_push(u32_t(core.math_coprocessor.registers.pop().value & 0xFFFFFFFF).value)

class SAVEW(Descriptor_MATH):
  """
  Downsize TOS to ``int``, and store the result in register.
  """

  mnemonic = 'savew'
  opcode = MathCoprocessorOpcodes.SAVEW
  operands = 'r'

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg1', 5, operands['register_n0'])

  @staticmethod
  def disassemble_operands(logger, inst):
    return [REGISTER_NAMES[inst.reg1]]

  @staticmethod
  def execute(core, inst):
    core.registers.map[inst.reg1].value = u32_t(core.math_coprocessor.registers.pop().value & 0xFFFFFFFF).value

class POPW(Descriptor_MATH):
  """
  Pop the ``int`` from data stack, extend it to ``long``, and make the result TOS.
  """

  mnemonic = 'popw'
  opcode = MathCoprocessorOpcodes.POPW

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.sign_extend_with_push(core.raw_pop())

class LOADW(Descriptor_MATH):
  """
  Take a value from register, extend it to ``long``, and make the result TOS.
  """

  mnemonic = 'loadw'
  opcode = MathCoprocessorOpcodes.LOADW
  operands = 'r'

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg1', 5, operands['register_n0'])

  @staticmethod
  def disassemble_operands(logger, inst):
    return [REGISTER_NAMES[inst.reg1]]

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.sign_extend_with_push(core.registers.map[inst.reg1].value)

class POPUW(Descriptor_MATH):
  """
  Pop the ``int``from data stack, extend it to ``long``, and make the result TOS.
  """

  mnemonic = 'popuw'
  opcode = MathCoprocessorOpcodes.POPUW

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.extend_with_push(core.raw_pop())

class LOADUW(Descriptor_MATH):
  """
  Take a value from register, extend it to ``long``, and make the result TOS.
  """

  mnemonic = 'loaduw'
  opcode = MathCoprocessorOpcodes.LOADUW
  operands = 'r'

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg1', 5, operands['register_n0'])

  @staticmethod
  def disassemble_operands(logger, inst):
    return [REGISTER_NAMES[inst.reg1]]

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.extend_with_push(core.registers.map[inst.reg1].value)

class PUSH(Descriptor_MATH):
  """
  Push the TOS on the data stack.
  """

  mnemonic = 'push'
  opcode = MathCoprocessorOpcodes.PUSH

  @staticmethod
  def execute(core, inst):
    v = core.math_coprocessor.registers.pop()

    core.raw_push(v.value & 0xFFFFFFFF)
    core.raw_push(v.value >> 32)

class SAVE(Descriptor_MATH):
  """
  Store TOS in two registers.
  """

  mnemonic = 'save'
  opcode = MathCoprocessorOpcodes.SAVE
  operands = 'r,r'

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg1', 5, operands['register_n0'])
    ENCODE(logger, buffer, inst, 'reg2', 5, operands['register_n1'])

  @staticmethod
  def disassemble_operands(logger, inst):
    return [REGISTER_NAMES[inst.reg1], REGISTER_NAMES[inst.reg1]]

  @staticmethod
  def execute(core, inst):
    v = core.math_coprocessor.registers.pop().value

    core.registers.map[inst.reg1].value = v >> 32
    core.registers.map[inst.reg2].value = v & 0xFFFFFFFF

class POP(Descriptor_MATH):
  """
  Pop the ``long`` from data stack, and make it new TOS.
  """

  mnemonic = 'pop'
  opcode = MathCoprocessorOpcodes.POP

  @staticmethod
  def execute(core, inst):
    hi = core.raw_pop()
    lo = core.raw_pop()

    core.math_coprocessor.registers.push(u64_t((hi << 32) | lo))

class LOAD(Descriptor_MATH):
  """
  Merge two registers together, and make the result new TOS.
  """

  mnemonic = 'load'
  opcode = MathCoprocessorOpcodes.LOAD
  operands = 'r,r'

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg1', 5, operands['register_n0'])
    ENCODE(logger, buffer, inst, 'reg2', 5, operands['register_n1'])

  @staticmethod
  def disassemble_operands(logger, inst):
    return [REGISTER_NAMES[inst.reg1], REGISTER_NAMES[inst.reg1]]

  @staticmethod
  def execute(core, inst):
    hi = core.registers.map[inst.reg1].value
    lo = core.registers.map[inst.reg2].value

    core.math_coprocessor.registers.push(u64_t((hi << 32) | lo))

class INCL(Descriptor_MATH):
  """
  Increment top of the stack by one.
  """

  mnemonic = 'incl'
  opcode = MathCoprocessorOpcodes.INCL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.registers.tos().value += 1

class DECL(Descriptor_MATH):
  """
  Decrement top of the stack by one.
  """

  mnemonic = 'decl'
  opcode = MathCoprocessorOpcodes.DECL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.registers.tos().value -= 1

class ADDL(Descriptor_MATH):
  mnemonic = 'addl'
  opcode = MathCoprocessorOpcodes.ADDL

  @staticmethod
  def execute(core, inst):
    RS = core.math_coprocessor.registers

    a = RS.pop()
    b = RS.pop()

    RS.push(u64_t(a.value + b.value))

class MULL(Descriptor_MATH):
  """
  Multiply two top-most numbers on the stack.
  """

  mnemonic = 'mull'
  opcode = MathCoprocessorOpcodes.MULL

  @staticmethod
  def execute(core, inst):
    RS = core.math_coprocessor.registers

    a = RS.pop()
    b = RS.pop()

    RS.push(u64_t(a.value * b.value))

class DIVL(Descriptor_MATH):
  """
  Divide the value below the top of the math stack by the topmost value.
  """

  mnemonic = 'divl'
  opcode = MathCoprocessorOpcodes.DIVL

  @staticmethod
  def execute(core, inst):
    RS = core.math_coprocessor.registers

    divider = i64_t(RS.pop().value)
    tos = i64_t(RS.pop().value)

    RS.push(u64_t(tos.value // divider.value))

class MODL(Descriptor_MATH):
  mnemonic = 'modl'
  opcode = MathCoprocessorOpcodes.MODL

  @staticmethod
  def execute(core, inst):
    RS = core.math_coprocessor.registers

    divider = i64_t(RS.pop().value)
    tos = i64_t(RS.pop().value)

    RS.push(u64_t(tos.value % divider.value))

class UDIVL(Descriptor_MATH):
  """
  Divide the value below the top of the math stack by the topmost value.
  """

  mnemonic = 'udivl'
  opcode = MathCoprocessorOpcodes.UDIVL

  @staticmethod
  def execute(core, inst):
    RS = core.math_coprocessor.registers

    divider = RS.pop()
    tos = RS.pop()

    RS.push(u64_t(tos.value // divider.value))

class UMODL(Descriptor_MATH):
  mnemonic = 'umodl'
  opcode = MathCoprocessorOpcodes.UMODL

  @staticmethod
  def execute(core, inst):
    RS = core.math_coprocessor.registers

    divider = RS.pop()
    tos = RS.pop()

    RS.push(u64_t(tos.value % divider.value))

class DUP(Descriptor_MATH):
  mnemonic = 'dup'
  opcode = MathCoprocessorOpcodes.DUP

  @staticmethod
  def execute(core, inst):
    M = core.math_coprocessor

    M.registers.push(u64_t(M.registers.tos().value))

class DUP2(Descriptor_MATH):
  mnemonic = 'dup2'
  opcode = MathCoprocessorOpcodes.DUP2

  @staticmethod
  def execute(core, inst):
    M = core.math_coprocessor

    a = M.registers.pop()
    b = M.registers.pop()
    M.registers.push(b)
    M.registers.push(a)
    M.registers.push(u64_t(b.value))
    M.registers.push(u64_t(a.value))

class SWP(Descriptor_MATH):
  mnemonic = 'swp'
  opcode = MathCoprocessorOpcodes.SWP

  @staticmethod
  def execute(core, inst):
    M = core.math_coprocessor

    a = M.registers.pop()
    b = M.registers.pop()
    M.registers.push(a)
    M.registers.push(b)

class DROP(Descriptor_MATH):
  mnemonic = 'drop'
  opcode = MathCoprocessorOpcodes.DROP

  @staticmethod
  def execute(core, inst):
    M = core.math_coprocessor

    M.registers.pop()

class SYMDIVL(Descriptor_MATH):
  """
  The same operation like ``DIVL`` but provides symmetric results.
  """

  mnemonic = 'symdivl'
  opcode = MathCoprocessorOpcodes.SYMDIVL

  @staticmethod
  def execute(core, inst):
    RS = core.math_coprocessor.registers

    divider = i64_t(RS.pop().value)
    tos = i64_t(RS.pop().value)

    RS.push(u64_t(int(float(tos.value) / float(divider.value))))

class SYMMODL(Descriptor_MATH):
  mnemonic = 'symmodl'
  opcode = MathCoprocessorOpcodes.SYMMODL

  @staticmethod
  def execute(core, inst):
    import math

    RS = core.math_coprocessor.registers

    divider = i64_t(RS.pop().value).value
    tos = i64_t(RS.pop().value).value

    if (tos < 0) == (divider < 0):
      RS.push(u64_t(tos % divider))

    else:
      RS.push(u64_t(int(math.fmod(tos, divider))))

ADDL(MathCoprocessorInstructionSet)
INCL(MathCoprocessorInstructionSet)
DECL(MathCoprocessorInstructionSet)
MULL(MathCoprocessorInstructionSet)
DIVL(MathCoprocessorInstructionSet)
MODL(MathCoprocessorInstructionSet)
UDIVL(MathCoprocessorInstructionSet)
UMODL(MathCoprocessorInstructionSet)
SYMDIVL(MathCoprocessorInstructionSet)
SYMMODL(MathCoprocessorInstructionSet)
DUP(MathCoprocessorInstructionSet)
DUP2(MathCoprocessorInstructionSet)
SWP(MathCoprocessorInstructionSet)
DROP(MathCoprocessorInstructionSet)
PUSHW(MathCoprocessorInstructionSet)
SAVEW(MathCoprocessorInstructionSet)
POPW(MathCoprocessorInstructionSet)
LOADW(MathCoprocessorInstructionSet)
POPUW(MathCoprocessorInstructionSet)
LOADUW(MathCoprocessorInstructionSet)
PUSH(MathCoprocessorInstructionSet)
SAVE(MathCoprocessorInstructionSet)
POP(MathCoprocessorInstructionSet)
LOAD(MathCoprocessorInstructionSet)
SIS(MathCoprocessorInstructionSet)

MathCoprocessorInstructionSet.init()

INSTRUCTION_SETS[MathCoprocessorInstructionSet.instruction_set_id] = MathCoprocessorInstructionSet



from six import itervalues

from . import BlockVisitor

class BlockTreeVisualiseVisitor(BlockVisitor):
  priority = 1000

  def __init__(self, *args, **kwargs):
    super(BlockTreeVisualiseVisitor, self).__init__(*args, **kwargs)

    self.diag = None

  def visit(self, cv):
    self.DEBUG(self.log_prefix + 'BT visualise pass')
    self.DOWN()

    self.diag = [
      'blockdiag {',
      'orientation = portrait;',
      'node_width = 256;',
      'node_height = 60;',
    ]

    self.do_visit(cv)

    self.diag.append('}')

    with open('bt.diag', 'w') as f:
      f.write('\n'.join(self.diag))

    self.UP()

  def do_visit_block(self, block):
    name = []

    if block.comment is not None:
      name.append(block.comment)

    if block.names:
      name.append('(%s)' % ', '.join(block.names))

    self.diag.append('%i [numbered = %i, label = "%s"];' % (block.id, block.id, ' '.join(name)))

    if block.outgoing:
      for out_block in itervalues(block.outgoing):
        self.diag.append('%i -> %i;' % (block.id, out_block.id))

from . import BT_PASSES
BT_PASSES['bt-visualise'] = BlockTreeVisualiseVisitor



import collections
import ctypes
import functools
import mmap
import os.path
import re

from six import iteritems, itervalues, integer_types, string_types, PY2

from .. import cpu
from .. import mm
from ..cpu.coprocessor.math_copro import MathCoprocessorInstructionSet  # noqa - it's not unused, SIS instruction may need it but that's hidden from flake
from ..cpu.instructions import encoding_to_u32

from ..mm import u8_t, PAGE_SIZE, UINT32_FMT
from ..mm.binary import SectionTypes, SectionFlags, SymbolFlags, RelocFlags
from ..util import align, str2bytes
from ..errors import AssemblerError, IncompleteDirectiveError, UnknownFileError, DisassembleMismatchError, UnknownPatternError, TooManyLabelsError

align_to_next_page = functools.partial(align, PAGE_SIZE)
align_to_next_mmap = functools.partial(align, mmap.PAGESIZE)

def PATTERN(pattern):
  return re.compile(r'^\s*(?P<payload>' + pattern + r')(?:\s*[;#].*)?$', re.MULTILINE)

RE_INTEGER = re.compile(r'^\s+(?:(?P<value_hex>-?0x[a-fA-F0-9]+)|(?P<value_dec>0|(?:-?[1-9][0-9]*))|(?P<value_var>[a-zA-Z][a-zA-Z0-9_]*(?:.*?)?)|(?P<value_label>&[a-zA-Z_\.][a-zA-Z0-9_\.]*))\s*$')
RE_STR = re.compile(r'^\s*"(?P<string>.*?)"\s*$')

RE_COMMENT = re.compile(r'^\s*[/;].*?$', re.MULTILINE)
RE_INCLUDE = PATTERN(r'\.include\s+"(?P<file>[a-zA-Z0-9_\-/\.]+)\s*"')
RE_IFDEF = PATTERN(r'.ifdef\s+(?P<var>[a-zA-Z0-9_]+)$')
RE_IFNDEF = PATTERN(r'\.ifndef\s+(?P<var>[a-zA-Z0-9_]+)')
RE_ELSE = PATTERN(r'\.else')
RE_ENDIF = PATTERN(r'\.endif')
RE_VAR_DEF = PATTERN(r'\.def\s+(?P<var_name>[a-zA-Z][a-zA-Z0-9_]*):\s*(?P<var_body>.*?)')
RE_MACRO_DEF = re.compile(r'^\s*\.macro\s+(?P<macro_name>[a-zA-Z_][a-zA-Z0-9_]*)(?:\s+(?P<macro_params>.*?))?:$', re.MULTILINE | re.DOTALL)
RE_MACRO_END = PATTERN(r'\.end')
RE_ASCII = PATTERN(r'\.ascii\s+(?P<string>".*?")')
RE_BYTE = PATTERN(r'\.byte(?P<integer>.*?)')
RE_DATA = PATTERN(r'\.data(?:\s+(?P<name>\.[a-z][a-z0-9_]*))?')
RE_INT = PATTERN(r'\.int(?P<integer>.*?)')
RE_LABEL = PATTERN(r'(?P<label>[a-zA-Z_\.][a-zA-Z0-9_\.]*):')
RE_SECTION = PATTERN(r'\.section\s+(?P<name>\.[a-zA-z0-9_]+)(?:,\s*(?P<flags>[rwxlbmg]*))?')
RE_SET = PATTERN(r'\.set\s+(?P<name>[a-zA-Z_][a-zA-Z0-9_]*),\s*(?:(?P<current>\.)|(?P<value_hex>-?0x[a-fA-F0-9]+)|(?P<value_dec>0|(?:-?[1-9][0-9]*))|(?P<value_label>&[a-zA-Z][a-zA-Z0-9_]*))')
RE_SHORT = PATTERN(r'\.short(?P<integer>.*?)')
RE_SIZE = PATTERN(r'\.size\s+(?P<size>[1-9][0-9]*)')
RE_SPACE = PATTERN(r'\.space\s+(?P<size>[1-9][0-9]*)')
RE_STRING = PATTERN(r'\.string(?P<string>.*?)')
RE_TEXT = PATTERN(r'\.text(?:\s+(?P<name>\.[a-z][a-z0-9_]*))?')
RE_TYPE = PATTERN(r'\.type\s+(?P<name>[a-zA-Z_\.][a-zA-Z0-9_]*),\s*(?P<type>(?:char|byte|short|int|ascii|string|space))')
RE_GLOBAL = PATTERN(r'\.global\s+(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)')
RE_ALIGN  = PATTERN(r'\.align\s+(?P<boundary>[0-9]+)')

class SourceLocation(object):
  def __init__(self, filename = None, lineno = None, column = None, length = None):
    self.filename = filename
    self.lineno = lineno
    self.column = column
    self.length = length

  def copy(self):
    return SourceLocation(filename = self.filename, lineno = self.lineno, column = self.column, length = self.length)

  def __str__(self):
    t = [self.filename, str(self.lineno)]
    if self.column is not None:
      t.append(str(self.column))
    return ':'.join(t)

  def __repr__(self):
    return str(self)

class Buffer(object):
  def __init__(self, logger, filename, buff):
    super(Buffer, self).__init__()

    self.logger = logger
    self.DEBUG = logger.debug
    self.INFO = logger.info
    self.WARN = logger.warning
    self.ERROR = logger.error
    self.EXCEPTION = logger.exception

    self.buff = buff

    self.location = SourceLocation(filename = filename, lineno = 0)
    self.last_line = None

  def get_line(self):
    while self.buff:
      self.location.lineno += 1

      line = self.buff.pop(0)

      if isinstance(line, SourceLocation):
        self.location = line

        self.DEBUG('buffer: file switch: %s', str(self.location))
        continue

      if not line:
        continue

      self.DEBUG('buffer: new line %s: %s', str(self.location), line)
      self.last_line = line
      return line

    self.last_line = None
    return None

  def put_line(self, line):
    self.buff.insert(0, line)
    self.location.lineno -= 1

  def put_buffer(self, buff, filename = None):
    filename = filename or '<unknown>'

    self.buff.insert(0, self.location.copy())

    if isinstance(buff, string_types):
      buff = buff.split('\n')

    for line in reversed(buff):
      self.buff.insert(0, line)

    self.buff.insert(0, SourceLocation(filename = filename, lineno = 0))

  def has_lines(self):
    return len(self.buff) > 0

  def get_error(self, cls, info, column = None, length = None, **kwargs):
    location = self.location.copy()
    location.column = column
    location.length = length

    kwargs['location'] = location
    if 'line' not in kwargs:
      kwargs['line'] = self.last_line

    kwargs['info'] = info

    return cls(**kwargs)

class Section(object):
  def __init__(self, s_name, s_type, s_flags):
    super(Section, self).__init__()

    self.name    = s_name
    self.type    = s_type
    self.flags   = s_flags
    self.content = []

    self.base = None
    self.ptr  = 0

  def __getattr__(self, name):
    if name == 'data_size':
      return sum([sizeof(i) for i in self.content])

    if name == 'file_size':
      return align_to_next_mmap(self.data_size) if self.flags.mmapable == 1 else self.data_size

    if name == 'items':
      return len(self.content)

  def __repr__(self):
    return '<Section: name=%s, type=%s, flags=%s, base=%s, ptr=%s, items=%s, data_size=%s, file_size=%s>' % (self.name, self.type, self.flags.to_string(), UINT32_FMT(self.base), UINT32_FMT(self.ptr), self.items, self.data_size, self.file_size)

class TextSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(TextSection, self).__init__(s_name, SectionTypes.TEXT, flags or SectionFlags.create(readable = True, executable = True, loadable = True))

class RODataSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(RODataSection, self).__init__(s_name, SectionTypes.DATA, flags or SectionFlags.create(readable = True, loadable = True))

class DataSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(DataSection, self).__init__(s_name, SectionTypes.DATA, flags or SectionFlags.create(readable = True, writable = True, loadable = True))

class BssSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(BssSection, self).__init__(s_name, SectionTypes.DATA, flags or SectionFlags.create(readable = True, writable = True, loadable = True, bss = True))

class SymbolsSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(SymbolsSection, self).__init__(s_name, SectionTypes.SYMBOLS, SectionFlags.create())

class RelocSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(RelocSection, self).__init__(s_name, SectionTypes.RELOC, SectionFlags.create())

class Label(object):
  def __init__(self, name, section, location):
    super(Label, self).__init__()

    self.name = name
    self.section = section
    self.location = location

  def __repr__(self):
    return '<label {} in section {} ({})>'.format(self.name, self.section.name if self.section else None, str(self.location))

class Reference(object):
  def __init__(self, add = None, label = None):
    self.add = add or 0
    self.label = label

  def __repr__(self):
    return '<Reference: label=%s, add=%s>' % (self.label, self.add)

class RelocSlot(object):
  def __init__(self, name, flags = None, patch_section = None, patch_address = None, patch_offset = None, patch_size = None, patch_add = None):
    super(RelocSlot, self).__init__()

    self.name = name
    self.flags = flags or RelocFlags.create()
    self.patch_section = patch_section
    self.patch_address = patch_address
    self.patch_offset = patch_offset
    self.patch_size = patch_size
    self.patch_add = patch_add

    self.size = 0

  def __repr__(self):
    return '<RelocSlot: name=%s, flags=%s, section=%s, address=%s, offset=%s, size=%s, add=%s>' % (self.name, self.flags.to_string(), self.patch_section, UINT32_FMT(self.patch_address), self.patch_offset, self.patch_size, self.patch_add)

class DataSlot(object):
  def __init__(self):
    super(DataSlot, self).__init__()

    self.name  = None
    self.size  = None
    self.refers_to = None
    self.value = None

    self.flags = SymbolFlags.create()

    self.section = None
    self.section_ptr = None

    self.location = None

  def close(self):
    pass

class ByteSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.CHAR

  def close(self):
    self.size = 1

    if self.refers_to:
      return

    self.value = [u8_t(self.value or 0)]

  def __repr__(self):
    return '<ByteSlot: name={}, size={}, section={}, value={}>'.format(self.name, self.size, self.section.name if self.section else '', self.value)

class ShortSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.SHORT

  def close(self):
    self.size = 2

    if self.refers_to:
      return

    value = self.value or 0
    self.value = [u8_t(value), u8_t(value >> 8)]

  def __repr__(self):
    return '<ShortSlot: name={}, size={}, section={}, value={}, refers_to={}>'.format(self.name, self.size, self.section.name if self.section else '', self.value, self.refers_to)

class IntSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.INT

  def close(self):
    self.size = 4

    if self.refers_to:
      return

    value = self.value or 0
    self.value = [u8_t(value), u8_t(value >> 8), u8_t(value >> 16), u8_t(value >> 24)]

  def __repr__(self):
    return '<IntSlot: name={}, size={}, section={}, value={}, refers_to={}>'.format(self.name, self.size, self.section.name if self.section else '', self.value, self.refers_to)

class CharSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.CHAR

  def close(self):
    self.size = 1
    self.value = u8_t(ord(self.value or '\0'))

  def __repr__(self):
    return '<CharSlot: name={}, section={}, value={}>'.format(self.name, self.section.name if self.section else '', self.value)

class SpaceSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.ASCII

  def close(self):
    self.value = None
    self.size = self.size

  def __repr__(self):
    return '<SpaceSlot: name={}, size={}, section={}>'.format(self.name, self.size, self.section.name if self.section else '')

class AsciiSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.ASCII

  def close(self):
    self.value = self.value or ''
    self.value = [u8_t(ord(c)) for c in self.value]
    self.size = len(self.value)

  def __repr__(self):
    return '<AsciiSlot: name={}, size={}, section={}, value={}>'.format(self.name, self.size, self.section.name if self.section else '', self.value)

class StringSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.STRING

  def close(self):
    self.value = self.value or ''
    self.value = [u8_t(ord(c)) for c in self.value] + [u8_t(0)]
    self.size = len(self.value)

  def __repr__(self):
    return '<StringSlot: name={}, size={}, section={}, value={}>'.format(self.name, self.size, self.section.name if self.section else '', self.value)

class BytesSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.ASCII

  def close(self):
    self.value = self.value or ''
    self.value = [u8_t(b) for b in self.value]
    self.size = len(self.value)

  def __repr__(self):
    return '<BytesSlot: name={}, size={}, section={}, value={}>'.format(self.name, self.size, self.section.name if self.section else '', self.value)

class AlignSlot(DataSlot):
  def __init__(self, boundary):
    super(AlignSlot, self).__init__()

    self.boundary = boundary

  def __repr__(self):
    return '<AlignSlot: boundary={}>'.format(self.boundary)

class FunctionSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.FUNCTION

  def close(self):
    self.size = 0

  def __repr__(self):
    return '<FunctionSlot: name={}, section={}>'.format(self.name, self.section.name if self.section else '')

def sizeof(o):
  if isinstance(o, RelocSlot):
    return 0

  if isinstance(o, DataSlot):
    return o.size

  if isinstance(o, ctypes.LittleEndianStructure):
    return ctypes.sizeof(o)

  return ctypes.sizeof(o)

if PY2:
  def decode_string(s):
    return s.decode('string_escape')

else:
  def decode_string(s):
    return str2bytes(s).decode('unicode_escape')

def translate_buffer(logger, buff, base_address = None, mmapable_sections = False, writable_sections = False, filename = None, defines = None, includes = None, verify_disassemble = False):
  DEBUG = logger.debug

  base_address = base_address or 0
  filename = filename or '<unknown>'
  defines = defines or []
  includes = includes or []
  includes.insert(0, os.getcwd())

  defines = {var[0]: var[1] if len(var) > 1 else None for var in (var.split('=') for var in defines)}

  DEBUG('translate_buffer: base_addres=%s, mmapable_sections=%s, writable_sections=%s, filename=%s, defines=%s, includes=%s, verify_disassemble=%s', UINT32_FMT(base_address), mmapable_sections, writable_sections, filename, defines, includes, verify_disassemble)

  buff = Buffer(logger, filename, buff.split('\n'))

  sections_pass1 = collections.OrderedDict([
    ('.text',   TextSection('.text')),
    ('.rodata', RODataSection('.rodata')),
    ('.data',   DataSection('.data')),
    ('.bss',    BssSection('.bss')),
    ('.symtab', SymbolsSection('.symtab')),
    ('.reloc',  RelocSection('.reloc'))
  ])

  if mmapable_sections:
    for section in itervalues(sections_pass1):
      section.flags.mmapable = True

  if writable_sections:
    for section in [_section for _section in itervalues(sections_pass1) if _section.name in ('.text', '.rodata', '.data', '.bss')]:
      section.flags.writable = True

  DEBUG('Pass #1')

  labeled = []

  line = None

  def __apply_defs(line):
    orig_line = line

    for def_pattern, def_value in iteritems(defs):
      line = def_pattern.sub(def_value, line)

    if orig_line != line:
      DEBUG(msg_prefix + 'variables replaced: line="%s"', line)

    return line

  def __apply_macros(line):
    for m_pattern, m_desc in iteritems(macros):
      matches = m_pattern.match(line)
      if not matches:
        continue

      DEBUG(msg_prefix + 'replacing macro: name=%s', m_desc['name'])

      if m_desc['params']:
        matches = matches.groupdict()

        replace_map = {}
        for i in range(0, len(m_desc['params'])):
          replace_map[re.compile(r'#{}'.format(m_desc['params'][i]))] = matches['arg{}'.format(i)]

        DEBUG(msg_prefix + 'macro args: %s', ', '.join(['{} => {}'.format(pattern.pattern, repl) for pattern, repl in iteritems(replace_map)]))

        body = []
        for line in m_desc['body']:
          for pattern, repl in iteritems(replace_map):
            line = pattern.sub(repl, line)
          body.append(line)

        buff.put_buffer(body)

      else:
        buff.put_buffer(m_desc['body'])

      return True

    return False

  def __set_var_location(var, location = None):
    if var.location is None:
      if location is None:
        location = buff.location.copy()

      var.location = location

  def __set_var_label(var):
    if len(labels) > 1:
      raise buff.get_error(TooManyLabelsError, 'Too many labels', column = 0)

    var.name = labels[0] if labels else None

  def __parse_integer(var, matches, max):
    __set_var_location(var)

    groupdict = matches.groupdict()
    integer = groupdict.get('integer')

    DEBUG('__parse_integer: var=%s, max=%s, matches=%s', var, max, groupdict)

    if integer is None or not integer:
      raise buff.get_error(IncompleteDirectiveError, 'directive without a value specification', column = matches.end(1))

    integer_start = matches.start(2)

    matches = RE_INTEGER.match(integer)
    if matches is None:
      raise buff.get_error(IncompleteDirectiveError, 'directive without a meaningful value', column = integer_start)

    groupdict = matches.groupdict()
    DEBUG('__parse_integer: matches=%s', groupdict)

    v_value = groupdict.get('value_dec')
    if v_value:
      var.value = int(v_value)

      DEBUG('__parse_integer: var=%s', var)
      return

    v_value = groupdict.get('value_hex')
    if v_value:
      var.value = int(v_value, base = 16)

      DEBUG('__parse_integer: var=%s', var)
      return

    v_value = groupdict.get('value_var')
    if v_value:
      if v_value not in variables:
        raise buff.get_error(IncompleteDirectiveError, 'unknown variable named "%s"' % v_value, column = integer_start)

      variable = variables[v_value]

      DEBUG('__parse_integer: variable: variable=%s', variable)

      if isinstance(variable, integer_types):
        var.value = variable

      else:
        var.refers_to = Reference(label = variable)

      DEBUG('__parse_integer: var=%s', var)
      return

    v_value = groupdict.get('value_label')
    if v_value:
      var.refers_to = Reference(label = v_value)

      DEBUG('__parse_integer: var=%s', var)
      return

    raise buff.get_error(IncompleteDirectiveError, 'directive without a meaningful value', column = integer_start)

  def __parse_string(var, matches):
    __set_var_location(var)

    groupdict = matches.groupdict()

    string = groupdict.get('string')

    DEBUG('__parse_string: var=%s, max=%s, matches=%s', var, max, groupdict)

    if string is None or not string:
      raise buff.get_error(IncompleteDirectiveError, 'directive without a value specification', column = matches.end(1))

    string_start = matches.start(2)

    matches = RE_STR.match(string)
    if matches is None:
      raise buff.get_error(IncompleteDirectiveError, 'directive without a meaningful value', column = string_start)

    groupdict = matches.groupdict()

    v_value = groupdict.get('string')
    if not v_value:
      raise buff.get_error(IncompleteDirectiveError, 'directive without a meaningful value', column = string_start)

    v_value = __apply_defs(v_value)

    DEBUG('Pre-decode: (%s) %s', type(v_value), ', '.join([str(ord(c)) for c in v_value]))
    var.value = decode_string(v_value)
    DEBUG('Post-decode: (%s) %s', type(var.value), ', '.join([str(ord(c)) for c in var.value]))

  def __parse_space(var, matches):
    __set_var_location(var)

    matches = matches.groupdict()

    if 'size' not in matches:
      raise buff.get_error(IncompleteDirectiveError, '.size directive without a size')

    var.size = int(matches['size'])

  def __handle_symbol_variable(v_name, v_type):
    if v_type == 'char':
      var = CharSlot()

    elif v_type == 'byte':
      var = ByteSlot()

    elif v_type == 'short':
      var = ShortSlot()

    elif v_type == 'int':
      var = IntSlot()

    elif v_type == 'ascii':
      var = AsciiSlot()

    elif v_type == 'string':
      var = StringSlot()

    elif v_type == 'space':
      var = SpaceSlot()

    var.name = Label(v_name, curr_section, buff.location.copy())
    __set_var_location(var)

    while buff.has_lines() and var.value is None and var.refers_to is None:
      line = buff.get_line()

      if line is None:
        var.close()
        data_section.content.append(var)
        return

      matches = RE_COMMENT.match(line)
      if matches:
        continue

      msg_prefix = 'pass #1: %s: ' % str(buff.location)

      line = __apply_defs(line)

      if not current_macro and __apply_macros(line):
        DEBUG(msg_prefix + 'macro replaced, get fresh line')
        continue

      matches = RE_TYPE.match(line)
      if matches:
        buff.put_line(line)
        break

      matches = RE_SIZE.match(line)
      if matches:
        matches = matches.groupdict()

        if 'size' not in matches:
          raise buff.get_error(IncompleteDirectiveError, '.size directive without a size')

        var.size = int(matches['size'])
        continue

      matches = RE_SHORT.match(line)
      if matches:
        __parse_integer(var, matches, 0xFFFF)
        continue

      matches = RE_INT.match(line)
      if matches:
        __parse_integer(var, matches, 0xFFFFFFFF)
        continue

      matches = RE_ASCII.match(line)
      if matches:
        __parse_string(var, matches)
        continue

      matches = RE_STRING.match(line)
      if matches:
        __parse_string(var, matches)
        continue

      matches = RE_SPACE.match(line)
      if matches:
        __parse_space(var, matches)
        continue

      matches = RE_BYTE.match(line)
      if matches:
        __parse_integer(var, matches, 0xFF)
        continue

      buff.put_line(line)
      break

    var.close()
    data_section.content.append(var)

  labels = []
  variables = {}

  instruction_set = cpu.instructions.DuckyInstructionSet

  defs = collections.OrderedDict()

  for name, value in iteritems(defines):
    if value is None:
      continue

    defs[re.compile(r'\${}'.format(name))] = value.strip()

  macros = collections.OrderedDict()
  current_macro = None

  DEBUG('Pass #1: text section is .text')
  DEBUG('Pass #1: data section is .data')

  text_section = sections_pass1['.text']
  data_section = sections_pass1['.data']
  curr_section = text_section

  global_symbols = []

  ifs = []

  def __fast_forward():
    DEBUG(msg_prefix + 'fast forwarding')

    depth = 1

    while buff.has_lines():
      line = buff.get_line()

      if line is None:
        return

      if not line.strip():
        continue

      matches = RE_IFDEF.match(line)
      if matches:
        depth += 1
        continue

      matches = RE_IFNDEF.match(line)
      if matches:
        depth += 1
        continue

      matches = RE_ENDIF.match(line)
      if matches:
        depth -= 1
        if depth == 0:
          buff.put_line(line)
          return

      matches = RE_ELSE.match(line)
      if matches:
        depth -= 1
        if depth == 0:
          buff.put_line(line)
          return

  while buff.has_lines():
    line = buff.get_line()

    if line is None:
      break

    if not line.strip():
      continue

    msg_prefix = 'pass #1: %s: ' % str(buff.location)

    line = __apply_defs(line)

    if not current_macro and __apply_macros(line):
      DEBUG(msg_prefix + 'macro replaced, get fresh line')
      continue

    matches = RE_COMMENT.match(line)
    if matches:
      continue

    msg_prefix = 'pass #1: %s: ' % str(buff.location)

    matches = RE_IFDEF.match(line)
    if matches:
      var = matches.groupdict()['var']

      DEBUG(msg_prefix + 'ifdef %s', var)

      ifs.append((True, var))

      if var in defines:
        DEBUG(msg_prefix + 'defined, continue processing')
        continue

      __fast_forward()
      continue

    matches = RE_IFNDEF.match(line)
    if matches:
      var = matches.groupdict()['var']

      DEBUG(msg_prefix + 'ifndef %s', var)

      ifs.append((False, var))

      if var not in defines:
        DEBUG(msg_prefix + 'not defined, continue processing')
        continue

      __fast_forward()
      continue

    matches = RE_ENDIF.match(line)
    if matches:
      DEBUG(msg_prefix + 'removing the last conditional from stack: %s', ifs[-1])

      ifs.pop()
      continue

    matches = RE_ELSE.match(line)
    if matches:
      defined, var = ifs.pop()

      DEBUG(msg_prefix + 'previous block was "%s %s"', 'ifdef' if defined is True else 'ifndef', var)

      ifs.append((not defined, var))

      if defined and var in defines:
        __fast_forward()
        continue

      DEBUG(msg_prefix + 'continue processing')
      continue

    matches = RE_INCLUDE.match(line)
    if matches:
      groupdict = matches.groupdict()

      if 'file' not in groupdict:
        raise buff.get_error(IncompleteDirectiveError, '.include directive without path', column = 0)

      DEBUG(msg_prefix + 'include: file=%s', groupdict['file'])

      replace = None

      for d in includes:
        filename = os.path.join(d, groupdict['file'])
        DEBUG(msg_prefix + '  checking file %s', filename)

        try:
          with open(filename, 'r') as f_in:
            replace = f_in.read()

        except IOError:
          DEBUG('    failed to read')
          pass  # "empty body on ExceptHandler" without this, because of patching

        else:
          DEBUG('    read as replacement')
          break

      if replace is None:
        raise buff.get_error(UnknownFileError, groupdict['file'], column = matches.start(2))

      buff.put_buffer(replace, filename = filename)

      continue

    matches = RE_VAR_DEF.match(line)
    if matches:
      matches = matches.groupdict()

      v_name = matches.get('var_name')
      v_body = matches.get('var_body')

      if not v_name or not v_body:
        raise buff.get_error(IncompleteDirectiveError, 'bad variable definition')

      DEBUG(msg_prefix + 'variable defined: name=%s, value=%s', v_name, v_body)

      defs[re.compile(r'\${}'.format(v_name))] = v_body.strip()

      continue

    matches = RE_MACRO_DEF.match(line)
    if matches:
      matches = matches.groupdict()

      m_name = matches.get('macro_name')
      m_params = matches.get('macro_params')

      if not m_name:
        raise buff.get_error(IncompleteDirectiveError, 'bad macro definition')

      DEBUG(msg_prefix + 'macro defined: name=%s', m_name)

      if current_macro:
        raise buff.get_error(AssemblerError, 'overlapping macro definitions')

      current_macro = {
        'name':    m_name,
        'pattern': None,
        'params':  [p.strip() for p in m_params.strip().split(',')] if m_params else [],
        'body':    []
      }

      if current_macro['params'] and len(current_macro['params'][0]):
        arg_pattern = r'(?P<arg%i>(?:".*?")|(?:.*?))'
        arg_patterns = ',\s*'.join([arg_pattern % i for i in range(0, len(current_macro['params']))])

        current_macro['pattern'] = re.compile(r'^\s*\${}\s+{}\s*(?:[;/#].*)?$'.format(m_name, arg_patterns), re.MULTILINE)

      else:
        current_macro['pattern'] = re.compile(r'\s*\${}'.format(m_name))

      continue

    matches = RE_MACRO_END.match(line)
    if matches:
      if not current_macro:
        raise buff.get_error(AssemblerError, 'closing non-existing macro')

      macros[current_macro['pattern']] = current_macro
      DEBUG(msg_prefix + 'macro definition closed: name=%s', current_macro['name'])
      current_macro = None
      continue

    if current_macro:
      current_macro['body'].append(line)
      continue

    matches = RE_SECTION.match(line)
    if matches:
      matches = matches.groupdict()

      if 'name' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.section directive without section name')

      s_name = matches['name']

      if s_name not in sections_pass1:
        section_flags = SectionFlags.from_string(matches.get('flags') or '')
        section_type = SectionTypes.TEXT if section_flags.executable is True else SectionTypes.DATA

        section = sections_pass1[s_name] = Section(s_name, section_type, section_flags)
        DEBUG(msg_prefix + 'section %s created', s_name)

      curr_section = sections_pass1[s_name]

      if curr_section.type == SectionTypes.TEXT:
        text_section = curr_section
        DEBUG(msg_prefix + 'text section changed to %s', s_name)
      else:
        data_section = curr_section
        DEBUG(msg_prefix + 'data section changed to %s', s_name)

      continue

    matches = RE_DATA.match(line)
    if matches:
      matches = matches.groupdict()

      curr_section = data_section = sections_pass1[matches['name'] if 'name' in matches and matches['name'] else '.data']
      DEBUG(msg_prefix + 'data section is %s', data_section.name)
      continue

    matches = RE_TEXT.match(line)
    if matches:
      matches = matches.groupdict()

      curr_section = text_section = sections_pass1[matches['name'] if 'name' in matches and matches['name'] else '.text']
      DEBUG(msg_prefix + 'text section is %s', text_section.name)
      continue

    matches = RE_TYPE.match(line)
    if matches:
      matches = matches.groupdict()

      if 'type' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.type directive without a type')

      if 'name' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.type directive without a name')

      __handle_symbol_variable(matches['name'], matches['type'])

      continue

    matches = RE_BYTE.match(line)
    if matches:
      var = ByteSlot()
      __parse_integer(var, matches, 0xFF)
      __set_var_label(var)

      var.close()

      DEBUG(msg_prefix + 'record byte value: name=%s, value=%s', var.name, var.value)
      data_section.content.append(var)

      labels = []
      continue

    matches = RE_SHORT.match(line)
    if matches:
      var = ShortSlot()
      __parse_integer(var, matches, 0xFFFF)
      __set_var_label(var)

      var.name = labels[0] if labels else None
      var.close()

      DEBUG(msg_prefix + 'record byte value: name=%s, value=%s', var.name, var.value)
      data_section.content.append(var)

      labels = []
      continue

    matches = RE_INT.match(line)
    if matches:
      var = IntSlot()
      __parse_integer(var, matches, 0xFFFFFFFF)
      __set_var_label(var)

      var.close()

      DEBUG(msg_prefix + 'record int value: name=%s, value=%s, refers_to=%s', var.name, var.value, var.refers_to)
      data_section.content.append(var)

      labels = []
      continue

    matches = RE_ASCII.match(line)
    if matches:
      var = AsciiSlot()
      __parse_string(var, matches)
      __set_var_label(var)

      var.close()

      DEBUG(msg_prefix + 'record ascii value: name=%s, value=%s', var.name, var.value)
      data_section.content.append(var)

      labels = []
      continue

    matches = RE_STRING.match(line)
    if matches:
      var = StringSlot()
      __parse_string(var, matches)
      __set_var_label(var)

      var.close()

      DEBUG(msg_prefix + 'record string value: name=%s, value=%s', var.name, var.value)
      data_section.content.append(var)

      labels = []
      continue

    matches = RE_SPACE.match(line)
    if matches:
      var = SpaceSlot()
      __parse_space(var, matches)
      __set_var_label(var)

      var.close()

      DEBUG(msg_prefix + 'record space: name=%s, value=%s', var.name, var.size)
      data_section.content.append(var)

      labels = []
      continue

    matches = RE_ALIGN.match(line)
    if matches:
      matches = matches.groupdict()

      if 'boundary' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.align directive without boundary')

      var = AlignSlot(int(matches['boundary']))

      DEBUG(msg_prefix + 'align: boundary=%s', var.boundary)
      data_section.content.append(var)
      continue

    matches = RE_GLOBAL.match(line)
    if matches:
      matches = matches.groupdict()

      if 'name' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.global directive without variable')

      name = matches['name']

      global_symbols.append(name)
      continue

    matches = RE_SET.match(line)
    if matches:
      matches = matches.groupdict()

      if 'name' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.set directive without variable')

      name = matches['name']

      if matches.get('current'):
        value = (curr_section.name, curr_section.ptr)

      elif matches.get('value_dec'):
        value = int(matches['value_dec'])

      elif matches.get('value_hex'):
        value = int(matches['value_hex'], base = 16)

      elif matches.get('value_label'):
        value = matches['value_label']

      else:
        raise buff.get_error(IncompleteDirectiveError, '.set directive with unknown value')

      DEBUG(msg_prefix + 'set variable: name=%s, value=%s', name, value)
      variables[name] = value

      continue

    matches = RE_LABEL.match(line)
    if matches:
      DEBUG('matches: %s', matches.groupdict())

      loc = buff.location.copy()
      loc.column = matches.start(2)

      label = Label(matches.groupdict()['label'], curr_section, loc)
      labels.append(label)

      DEBUG(msg_prefix + 'record label: name=%s', label.name)
      continue

    DEBUG(msg_prefix + 'line: %s', line)

    # label, instruction, 2nd pass flags
    emited_inst = None

    # Find instruction descriptor
    DEBUG(msg_prefix + 'instr set: %s', instruction_set)

    for desc in instruction_set.instructions:
      if not desc.pattern.match(line):
        continue
      break

    else:
      raise buff.get_error(UnknownPatternError, line, column = 0)

    emited_inst = desc.emit_instruction(logger, buff, line)
    emited_inst.desc = desc

    if labels:
      text_section.content.append((labels, emited_inst))

    else:
      text_section.content.append((None, emited_inst))

    labels = []

    emited_inst_disassemble = emited_inst.desc.instruction_set.disassemble_instruction(logger, emited_inst)
    DEBUG(msg_prefix + 'emitted instruction: %s (%s)', emited_inst_disassemble, UINT32_FMT(encoding_to_u32(emited_inst)))

    if verify_disassemble and line != emited_inst_disassemble:
      raise buff.get_error(DisassembleMismatchError, 'input="%s", emitted="%s"' % (line, emited_inst_disassemble))

    if isinstance(desc, cpu.instructions.SIS):
      DEBUG(msg_prefix + 'switching istruction set: inst_set=%s', emited_inst.immediate)

      instruction_set = cpu.instructions.get_instruction_set(emited_inst.immediate)

  for s_name, section in iteritems(sections_pass1):
    DEBUG('pass #1: section %s', s_name)

    if section.type == SectionTypes.TEXT:
      for labeled, inst in section.content:
        DEBUG('pass #1: inst=%s, labeled=%s', inst, labeled)

    else:
      for var in section.content:
        DEBUG('pass #1: %s', var)

  DEBUG('Pass #2')

  sections_pass2 = collections.OrderedDict()
  references = {}
  base_ptr = base_address

  for s_name, p1_section in iteritems(sections_pass1):
    section = sections_pass2[s_name] = Section(s_name, p1_section.type, p1_section.flags)

  symtab = sections_pass2['.symtab']
  reloctab = sections_pass2['.reloc']

  for s_name, section in iteritems(sections_pass2):
    p1_section = sections_pass1[s_name]

    section.base = base_ptr
    section.ptr  = base_ptr

    DEBUG('pass #2: section %s - base=%s', section.name, UINT32_FMT(section.base))

    if section.type == SectionTypes.SYMBOLS or section.type == SectionTypes.RELOC:
      continue

    if section.type == SectionTypes.DATA:
      for var in p1_section.content:
        ptr_prefix = 'pass #2: ' + UINT32_FMT(section.ptr) + ': '

        DEBUG(ptr_prefix + str(var))

        if isinstance(var, AlignSlot):
          aligned_ptr = align(var.boundary, section.ptr)
          padding_bytes = aligned_ptr - section.ptr

          if padding_bytes == 0:
            DEBUG(ptr_prefix + '  align %s to multiple of %s: aligned already, ignore', UINT32_FMT(section.ptr), var.boundary)
            continue

          padding = SpaceSlot()
          padding.size = padding_bytes

          DEBUG(ptr_prefix + '  align %s to multiple of %s: %s padding bytes => %s', UINT32_FMT(section.ptr), var.boundary, padding.size, UINT32_FMT(section.ptr + padding.size))
          DEBUG(ptr_prefix + '  %s', padding)

          section.content.append(padding)
          section.ptr += padding.size
          DEBUG(ptr_prefix + '  padding stored')

          continue

        if var.name:
          var.section = section
          var.section_ptr = section.ptr
          references['&' + var.name.name] = var

          symtab.content.append(var)

        if var.refers_to is not None:
          reference = var.refers_to

          DEBUG(ptr_prefix + '  refers to: %s', reference)

          if reference.label is not None:
            reloc = RelocSlot(reference.label[1:], patch_section = section, patch_address = section.ptr, patch_offset = 0, patch_size = 16, patch_add = reference.add)
            DEBUG(ptr_prefix + '  reloc slot created: %s', reloc)

          else:
            raise Exception()

          reloctab.content.append(reloc)
          var.refers_to = None

        if isinstance(var, IntSlot):
          if var.value is not None:
            section.content += var.value
            DEBUG(ptr_prefix + '  value stored')

          else:
            section.content.append(var)
            DEBUG(ptr_prefix + '  value missing - reserve space, fix in next pass')

          if var.size != 4:
            raise Exception()
          section.ptr += var.size

        elif isinstance(var, ShortSlot):
          if var.value is not None:
            section.content += var.value
            DEBUG(ptr_prefix + '  value stored')

          else:
            section.content.append(var)
            DEBUG(ptr_prefix + '  value missing - reserve space, fix in next pass')

          section.ptr += var.size

        elif isinstance(var, ByteSlot):
          if var.value is not None:
            section.content += var.value
            DEBUG(ptr_prefix + '  value stored')

          else:
            section.content.append(var)
            DEBUG(ptr_prefix + '  value missing - reserve space, fix in next pass')
          section.ptr += var.size

        elif type(var) == AsciiSlot or type(var) == StringSlot or isinstance(var, BytesSlot):
          section.content += var.value
          section.ptr += var.size
          DEBUG(ptr_prefix + '  value stored')

        elif type(var) == SpaceSlot:
          section.content.append(var)
          section.ptr += var.size
          DEBUG(ptr_prefix + '  value stored')

    if section.type == SectionTypes.TEXT:
      for labeled, inst in p1_section.content:
        ptr_prefix = 'pass #2: ' + UINT32_FMT(section.ptr) + ': '

        DEBUG(ptr_prefix + '%s (%s)', inst.desc.instruction_set.disassemble_instruction(logger, inst), UINT32_FMT(encoding_to_u32(inst)))

        inst.address = section.ptr

        if labeled:
          for label in labeled:
            var = FunctionSlot()
            var.name = label
            var.section = section
            var.section_ptr = section.ptr

            __set_var_location(var, location = label.location)

            var.close()

            symtab.content.append(var)

            references['&' + label.name] = var
            DEBUG(ptr_prefix + 'label entry "%s" created', label)

        if inst.desc.operands and ('i' in inst.desc.operands or 'j' in inst.desc.operands) and hasattr(inst, 'refers_to') and inst.refers_to is not None:
          DEBUG(ptr_prefix + 'refers to: label=%s, relative=%s, inst_aligned=%s', inst.refers_to, inst.desc.relative_address, inst.desc.inst_aligned)
          DEBUG(ptr_prefix + 'refers to: %s', inst.refers_to)

          reloc = RelocSlot(inst.refers_to.label[1:], flags = RelocFlags.create(relative = inst.desc.relative_address, inst_aligned = inst.desc.inst_aligned), patch_section = section, patch_address = section.ptr)
          inst.fill_reloc_slot(logger, inst, reloc)
          sections_pass2['.reloc'].content.append(reloc)

          if inst.refers_to in references:
            reloc.patch_section = references[inst.refers_to].section

          DEBUG(ptr_prefix + 'reloc slot created: %s', reloc)
          inst.refers_to = None

        section.content.append(inst)
        section.ptr += 4

    base_ptr = align_to_next_mmap(section.ptr) if mmapable_sections else align_to_next_page(section.ptr)

  DEBUG('Pass #3')

  sections_pass3 = {}

  for s_name, p2_section in iteritems(sections_pass2):
    section = Section(s_name, p2_section.type, p2_section.flags)
    sections_pass3[s_name] = section
    section.base = p2_section.base
    section.ptr  = section.base

  symtab = sections_pass3['.symtab']
  reloctab = sections_pass3['.reloc']

  for s_name, section in iteritems(sections_pass3):
    DEBUG('pass #3: section %s', section.name)

    p2_section = sections_pass2[s_name]

    if section.type == SectionTypes.SYMBOLS:
      symtab = section

    elif section.type == SectionTypes.RELOC:
      reloctab = section

    for item in p2_section.content:
      ptr_prefix = 'pass #3: ' + UINT32_FMT(section.ptr) + ': '

      if section.type == SectionTypes.SYMBOLS:
        if (type(item.name) is Label and item.name.name in global_symbols) or item.name in global_symbols:
          item.flags.globally_visible = True

      elif type(item) == IntSlot:
        if item.refers_to:
          reference, item.refers_to = item.refers_to, None

          DEBUG(ptr_prefix + 'refers to: %s', reference)

          reloc = RelocSlot(reference.label[1:], patch_section = references[item.refers_to].section, patch_address = section.ptr, patch_offset = 0, patch_size = 16, patch_add = reference.add)
          DEBUG(ptr_prefix + 'reloc slot created: %s', reloc)

          reloctab.content.append(reloc)

          item.value = 0x79797979

        item.close()
        item = item.value

      elif hasattr(item, 'refers_to') and item.refers_to:
        reference, item.refers_to = item.refers_to, None

        DEBUG(ptr_prefix + 'refers to: label=%s, relative=%s, inst_aligned=%s', reference, item.desc.relative_address, item.desc.inst_aligned)

        reloc = RelocSlot(reference.label[1:], flags = RelocFlags.create(relative = item.desc.relative_address, inst_aligned = item.desc.inst_aligned), patch_section = section, patch_address = section.ptr, patch_add = reference.add)
        item.fill_reloc_slot(logger, item, reloc)
        DEBUG(ptr_prefix + 'reloc slot created: %s', reloc)
        reloctab.content.append(reloc)

      DEBUG(ptr_prefix + str(item))

      if not isinstance(item, list):
        item = [item]

      for i in item:
        section.content.append(i)
        section.ptr += sizeof(i)

    DEBUG('pass #3: section %s finished: %s', section.name, section)

  DEBUG('Bytecode sections:')
  for s_name, section in iteritems(sections_pass3):
    DEBUG(str(section))

  DEBUG('Bytecode translation completed')

  return sections_pass3
